# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2011 OpenStack Foundation
# Copyright 2013 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import re
from unittest import mock
from urllib import parse

from oslo_utils import strutils

import novaclient
from novaclient import api_versions
from novaclient import client as base_client
from novaclient import exceptions
from novaclient.tests.unit import fakes
from novaclient.tests.unit import utils
from novaclient.v2 import client

# regex to compare callback to result of get_endpoint()
# checks version number (vX or vX.X where X is a number)
# and also checks if the id is on the end
ENDPOINT_RE = re.compile(
    r"^get_http:__nova_api:8774_v\d(_\d)?_\w{32}$")

# accepts formats like v2 or v2.1
ENDPOINT_TYPE_RE = re.compile(r"^v\d(\.\d)?$")

# accepts formats like v2 or v2_1
CALLBACK_RE = re.compile(r"^get_http:__nova_api:8774_v\d(_\d)?$")

# fake image uuids
FAKE_IMAGE_UUID_1 = 'c99d7632-bd66-4be9-aed5-3dd14b223a76'
FAKE_IMAGE_UUID_2 = 'f27f479a-ddda-419a-9bbc-d6b56b210161'
FAKE_IMAGE_UUID_SNAPSHOT = '555cae93-fb41-4145-9c52-f5b923538a26'
FAKE_IMAGE_UUID_SNAP_DEL = '55bb23af-97a4-4068-bdf8-f10c62880ddf'
FAKE_IMAGE_UUID_BACKUP = '2f87e889-41a4-4778-8553-83f5eea68c5d'

# fake request id
FAKE_REQUEST_ID = fakes.FAKE_REQUEST_ID
FAKE_REQUEST_ID_LIST = fakes.FAKE_REQUEST_ID_LIST
FAKE_RESPONSE_HEADERS = {'x-openstack-request-id': FAKE_REQUEST_ID}

FAKE_SERVICE_UUID_1 = '75e9eabc-ed3b-4f11-8bba-add1e7e7e2de'
FAKE_SERVICE_UUID_2 = '1f140183-c914-4ddf-8757-6df73028aa86'

SERVER_TOPOLOGY = {
    "nodes": [
        {
            "cpu_pinning": {
                "0": 0,
                "1": 5
            },
            "host_node": 0,
            "memory_mb": 1024,
            "siblings": [
                [
                    0,
                    1
                ]
            ],
            "vcpu_set": [
                0,
                1
            ]
        },
        {
            "cpu_pinning": {
                "2": 1,
                "3": 8
            },
            "host_node": 1,
            "memory_mb": 2048,
            "siblings": [
                [
                    2,
                    3
                ]
            ],
            "vcpu_set": [
                2,
                3
            ]
        }
    ],
    "pagesize_kb": 4
}


class FakeClient(fakes.FakeClient, client.Client):

    def __init__(self, api_version, *args, **kwargs):
        client.Client.__init__(self, username='username', password='password',
                               project_id='project_id', auth_url='auth_url',
                               extensions=kwargs.get('extensions'),
                               direct_use=False, api_version=api_version)
        self.client = FakeSessionClient(api_version=api_version, **kwargs)


class FakeSessionClient(base_client.SessionClient):

    def __init__(self, *args, **kwargs):

        self.callstack = []
        self.visited = []
        self.auth = mock.Mock()
        self.session = mock.Mock()
        self.service_type = 'service_type'
        self.service_name = None
        self.endpoint_override = None
        self.interface = None
        self.region_name = None
        self.version = None
        self.api_version = kwargs.get('api_version')
        self.auth.get_auth_ref.return_value.project_id = 'tenant_id'
        # determines which endpoint to return in get_endpoint()
        # NOTE(augustina): this is a hacky workaround, ultimately
        # we need to fix our whole mocking architecture (fixtures?)
        if 'endpoint_type' in kwargs:
            self.endpoint_type = kwargs['endpoint_type']
        else:
            self.endpoint_type = 'endpoint_type'
        self.logger = mock.MagicMock()

    def get_endpoint(self, **kwargs):
        # check if endpoint matches expected format (eg, v2.1)
        if (hasattr(self, 'endpoint_type') and
                ENDPOINT_TYPE_RE.search(self.endpoint_type)):
            return "http://nova-api:8774/%s/" % self.endpoint_type
        else:
            return (
                "http://nova-api:8774/v2.1/190a755eef2e4aac9f06aa6be9786385")

    def request(self, url, method, **kwargs):
        return self._cs_request(url, method, **kwargs)

    def _cs_request(self, url, method, **kwargs):
        # Check that certain things are called correctly
        if method in ['GET', 'DELETE']:
            assert 'body' not in kwargs
        elif method == 'PUT':
            assert 'body' in kwargs

        if url is not None:
            # Call the method
            args = parse.parse_qsl(parse.urlparse(url)[4])
            kwargs.update(args)
            munged_url = url.rsplit('?', 1)[0]
            munged_url = munged_url.strip('/').replace('/', '_')
            munged_url = munged_url.replace('.', '_')
            munged_url = munged_url.replace('-', '_')
            munged_url = munged_url.replace(' ', '_')
            munged_url = munged_url.replace('!', '_')
            munged_url = munged_url.replace('@', '_')
            munged_url = munged_url.replace('%20', '_')
            munged_url = munged_url.replace('%3A', '_')
            munged_url = munged_url.replace('%', '_')
            callback = "%s_%s" % (method.lower(), munged_url)

        if url is None or callback == "get_http:__nova_api:8774":
            # To get API version information, it is necessary to GET
            # a nova endpoint directly without "v2/<tenant-id>".
            callback = "get_versions"
        elif CALLBACK_RE.search(callback):
            callback = "get_current_version"
        elif ENDPOINT_RE.search(callback):
            # compare callback to result of get_endpoint()
            # NOTE(sdague): if we try to call a thing that doesn't
            # exist, just return a 404. This allows the stack to act
            # more like we'd expect when making REST calls.
            raise exceptions.NotFound('404')

        # Handle fake glance v2 requests
        v2_image = False
        if callback.startswith('get_v2_images'):
            v2_image = True
            callback = callback.replace('get_v2_', 'get_')

        simulate_pagination_next_links = [
            'get_os_simple_tenant_usage',
            'get_os_simple_tenant_usage_tenant_id',
        ]
        if callback in simulate_pagination_next_links:
            while callback in self.visited:
                callback += '_next'

        if not hasattr(self, callback):
            raise AssertionError('Called unknown API method: %s %s, '
                                 'expected fakes method name: %s' %
                                 (method, url, callback))

        # Note the call
        self.visited.append(callback)
        self.callstack.append((method, url, kwargs.get('body')))

        status, headers, body = getattr(self, callback)(**kwargs)

        # If we're dealing with a glance v2 image response, the response
        # isn't wrapped like the compute images API proxy is, so handle that.
        if body and v2_image and 'image' in body:
            body = body['image']

        r = utils.TestResponse({
            "status_code": status,
            "text": body,
            "headers": headers,
        })

        if status >= 400:
            raise exceptions.from_response(r, body, url, method)

        return r, body

    def get_versions(self):
        return (200, FAKE_RESPONSE_HEADERS, {
            "versions": [
                {"status": "SUPPORTED", "updated": "2011-01-21T11:33:21Z",
                 "links": [{"href": "http://nova-api:8774/v2/",
                            "rel": "self"}],
                 "min_version": "",
                 "version": "",
                 "id": "v2.0"},
                {"status": "CURRENT", "updated": "2013-07-23T11:33:21Z",
                 "links": [{"href": "http://nova-api:8774/v2.1/",
                            "rel": "self"}],
                 "min_version": novaclient.API_MIN_VERSION.get_string(),
                 "version": novaclient.API_MAX_VERSION.get_string(),
                 "id": "v2.1"}
            ]})

    def get_current_version(self):
        versions = {
            # v2 doesn't contain version or min_version fields
            "v2": {
                "version": {
                    "status": "SUPPORTED",
                    "updated": "2011-01-21T11:33:21Z",
                    "links": [{
                        "href": "http://nova-api:8774/v2/",
                        "rel": "self"
                    }],
                    "id": "v2.0"
                }
            },
            "v2.1": {
                "version": {
                    "status": "CURRENT",
                    "updated": "2013-07-23T11:33:21Z",
                    "links": [{
                        "href": "http://nova-api:8774/v2.1/",
                        "rel": "self"
                    }],
                    "min_version": novaclient.API_MIN_VERSION.get_string(),
                    "version": novaclient.API_MAX_VERSION.get_string(),
                    "id": "v2.1"
                }
            }
        }

        # if an endpoint_type that matches a version wasn't initialized,
        #  default to v2.1
        endpoint = 'v2.1'

        if hasattr(self, 'endpoint_type'):
            if ENDPOINT_TYPE_RE.search(self.endpoint_type):
                if self.endpoint_type in versions:
                    endpoint = self.endpoint_type
                else:
                    raise AssertionError(
                        "Unknown endpoint_type:%s" % self.endpoint_type)

        return (200, FAKE_RESPONSE_HEADERS, versions[endpoint])

    #
    # agents
    #

    def get_os_agents(self, **kw):
        hypervisor = kw.get('hypervisor', 'kvm')
        return (200, {}, {
            'agents':
                [{'hypervisor': hypervisor,
                  'os': 'win',
                  'architecture': 'x86',
                  'version': '7.0',
                  'url': 'xxx://xxxx/xxx/xxx',
                  'md5hash': 'add6bb58e139be103324d04d82d8f545',
                  'id': 1},
                 {'hypervisor': hypervisor,
                  'os': 'linux',
                  'architecture': 'x86',
                  'version': '16.0',
                  'url': 'xxx://xxxx/xxx/xxx1',
                  'md5hash': 'add6bb58e139be103324d04d82d8f546',
                  'id': 2}]})

    def post_os_agents(self, body):
        return (200, {}, {'agent': {
                          'url': '/xxx/xxx/xxx',
                          'hypervisor': body['agent']['hypervisor'],
                          'md5hash': 'add6bb58e139be103324d04d82d8f546',
                          'version': '7.0',
                          'architecture': 'x86',
                          'os': 'win',
                          'id': 1}})

    def delete_os_agents_1(self, **kw):
        return (202, {}, None)

    def put_os_agents_1(self, body, **kw):
        return (200, {}, {
            "agent": {"url": "/yyy/yyyy/yyyy",
                      "version": "8.0",
                      "md5hash": "add6bb58e139be103324d04d82d8f546",
                      'id': 1}})

    #
    # Limits
    #

    def get_limits(self, **kw):
        absolute = {
            "maxTotalRAMSize": 51200,
            "maxServerMeta": 5,
            "maxImageMeta": 5
        }
        # 2.57 removes injected_file* entries from the response.
        if self.api_version < api_versions.APIVersion('2.57'):
            absolute.update({"maxPersonality": 5, "maxPersonalitySize": 10240})
        return (200, {}, {"limits": {
            "rate": [
                {
                    "uri": "*",
                    "regex": ".*",
                    "limit": [
                        {
                            "value": 10,
                            "verb": "POST",
                            "remaining": 2,
                            "unit": "MINUTE",
                            "next-available": "2011-12-15T22:42:45Z"
                        },
                        {
                            "value": 10,
                            "verb": "PUT",
                            "remaining": 2,
                            "unit": "MINUTE",
                            "next-available": "2011-12-15T22:42:45Z"
                        },
                        {
                            "value": 100,
                            "verb": "DELETE",
                            "remaining": 100,
                            "unit": "MINUTE",
                            "next-available": "2011-12-15T22:42:45Z"
                        }
                    ]
                },
                {
                    "uri": "*/servers",
                    "regex": "^/servers",
                    "limit": [
                        {
                            "verb": "POST",
                            "value": 25,
                            "remaining": 24,
                            "unit": "DAY",
                            "next-available": "2011-12-15T22:42:45Z"
                        }
                    ]
                }
            ],
            "absolute": absolute,
        }})

    #
    # Servers
    #

    def get_servers(self, **kw):
        servers = {"servers": [
            {'id': '1234', 'name': 'sample-server'},
            {'id': '5678', 'name': 'sample-server2'},
            {'id': '9014', 'name': 'help'}
        ]}
        if self.api_version >= api_versions.APIVersion('2.69'):
            # include "partial results" from non-responsive part of
            # infrastructure.
            servers['servers'].append(
                {'id': '9015',
                 'status': "UNKNOWN",
                 "links": [
                     {
                         "href": "http://fake/v2.1/",
                         "rel": "self"
                     },
                     {
                         "href": "http://fake",
                         "rel": "bookmark"
                     }
                 ]}
            )
        return (200, {}, servers)

    def get_servers_detail(self, **kw):
        servers = {"servers": [
            {
                "id": '1234',
                "name": "sample-server",
                "image": {
                    "id": FAKE_IMAGE_UUID_2,
                    "name": "sample image",
                },
                "flavor": {
                    "id": 1,
                    "name": "256 MiB Server",
                },
                "hostId": "e4d909c290d0fb1ca068ffaddf22cbd0",
                "status": "BUILD",
                "progress": 60,
                "addresses": {
                    "public": [
                        {
                            "version": 4,
                            "addr": "1.2.3.4",
                        },
                        {
                            "version": 4,
                            "addr": "5.6.7.8",
                        }],
                    "private": [{
                        "version": 4,
                        "addr": "10.11.12.13",
                    }],
                },
                "metadata": {
                    "Server Label": "Web Head 1",
                    "Image Version": "2.1"
                },
                "OS-EXT-SRV-ATTR:host": "computenode1",
                "security_groups": [{
                    'id': 1, 'name': 'securitygroup1',
                    'description': 'FAKE_SECURITY_GROUP',
                    'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7'
                }],
                "OS-EXT-MOD:some_thing": "mod_some_thing_value",
            },
            {
                "id": '5678',
                "name": "sample-server2",
                "image": {
                    "id": FAKE_IMAGE_UUID_1,
                    "name": "sample image",
                },
                "flavor": {
                    "id": 1,
                    "name": "256 MiB Server",
                },
                "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                "status": "ACTIVE",
                "addresses": {
                    "public": [
                        {
                            "version": 4,
                            "addr": "4.5.6.7",
                        },
                        {
                            "version": 4,
                            "addr": "5.6.9.8",
                        }],
                    "private": [{
                        "version": 4,
                        "addr": "10.13.12.13",
                    }],
                },
                "metadata": {
                    "Server Label": "DB 1"
                },
                "OS-EXT-SRV-ATTR:host": "computenode2",
                "security_groups": [
                    {
                        'id': 1, 'name': 'securitygroup1',
                        'description': 'FAKE_SECURITY_GROUP',
                        'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7'
                    },
                    {
                        'id': 2, 'name': 'securitygroup2',
                        'description': 'ANOTHER_FAKE_SECURITY_GROUP',
                        'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7'
                    }],
            },
            {
                "id": '9012',
                "name": "sample-server3",
                "image": "",
                "flavor": {
                    "id": 1,
                    "name": "256 MiB Server",
                },
                "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                "status": "ACTIVE",
                "addresses": {
                    "public": [
                        {
                            "version": 4,
                            "addr": "4.5.6.7",
                        },
                        {
                            "version": 4,
                            "addr": "5.6.9.8",
                        }],
                    "private": [{
                        "version": 4,
                        "addr": "10.13.12.13",
                    }],
                },
                "metadata": {
                    "Server Label": "DB 1"
                }
            },
            {
                "id": '9013',
                "name": "sample-server4",
                "flavor": {
                    "id": '80645cf4-6ad3-410a-bbc8-6f3e1e291f51',
                },
                "image": {
                    "id": '3e861307-73a6-4d1f-8d68-f68b03223032',
                },
                "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                "status": "ACTIVE",
            },
            {
                "id": '9014',
                "name": "help",
                "flavor": {
                    "id": '80645cf4-6ad3-410a-bbc8-6f3e1e291f51',
                },
                "image": {
                    "id": '3e861307-73a6-4d1f-8d68-f68b03223032',
                },
                "hostId": "9e107d9d372bb6826bd81d3542a419d6",
                "status": "ACTIVE",
            },
        ]}
        if self.api_version >= api_versions.APIVersion('2.69'):
            # include "partial results" from non-responsive part of
            # infrastructure.
            servers['servers'].append(
                {
                    "id": "9015",
                    "status": "UNKNOWN",
                    "tenant_id": "6f70656e737461636b20342065766572",
                    "created": "2018-12-03T21:06:18Z",
                    "links": [
                        {
                            "href": "http://fake/v2.1/",
                            "rel": "self"
                        },
                        {
                            "href": "http://fake",
                            "rel": "bookmark"
                        }
                    ]
                }
            )
        return (200, {}, servers)

    def post_servers(self, body, **kw):
        assert set(body.keys()) <= set(['server', 'os:scheduler_hints'])
        fakes.assert_has_keys(
            body['server'],
            required=['name', 'imageRef', 'flavorRef'],
            optional=['metadata', 'personality'])
        if 'personality' in body['server']:
            for pfile in body['server']['personality']:
                fakes.assert_has_keys(pfile, required=['path', 'contents'])
        if body['server']['name'] == 'some-bad-server':
            return (202, {}, self.get_servers_1235()[2])
        else:
            return (202, {}, self.get_servers_1234()[2])

    def get_servers_1234(self, **kw):
        server = self.get_servers_detail()[2]['servers'][0]
        if self.api_version >= api_versions.APIVersion('2.71'):
            server.update(
                {'server_groups': ['a67359fb-d397-4697-88f1-f55e3ee7c499']})
        return (200, {}, {'server': server})

    def get_servers_1235(self, **kw):
        r = {'server': self.get_servers_detail()[2]['servers'][0]}
        r['server']['id'] = '1235'
        r['server']['status'] = 'error'
        r['server']['fault'] = {'message': 'something went wrong!'}
        return (200, {}, r)

    def get_servers_5678(self, **kw):
        r = {'server': self.get_servers_detail()[2]['servers'][1]}
        return (200, {}, r)

    def get_servers_9012(self, **kw):
        r = {'server': self.get_servers_detail()[2]['servers'][2]}
        return (200, {}, r)

    def get_servers_9013(self, **kw):
        r = {'server': self.get_servers_detail()[2]['servers'][3]}
        return (200, {}, r)

    def get_servers_9014(self, **kw):
        r = {'server': self.get_servers_detail()[2]['servers'][4]}
        return (200, {}, r)

    def get_servers_9015(self, **kw):
        r = {'server': self.get_servers_detail()[2]['servers'][5]}
        r['server']["OS-EXT-AZ:availability_zone"] = 'geneva'
        r['server']["OS-EXT-STS:power_state"] = 0
        flavor = {
            "disk": 1,
            "ephemeral": 0,
            "original_name": "m1.tiny",
            "ram": 512,
            "swap": 0,
            "vcpus": 1,
            "extra_specs": {}
        }
        image = {
            "id": "c99d7632-bd66-4be9-aed5-3dd14b223a76",
        }
        r['server']['image'] = image
        r['server']['flavor'] = flavor
        r['server']['user_id'] = "fake"
        return (200, {}, r)

    def delete_os_server_groups_12345(self, **kw):
        return (202, {}, None)

    def delete_os_server_groups_56789(self, **kw):
        return (202, {}, None)

    def delete_servers_1234(self, **kw):
        return (202, {}, None)

    def delete_servers_5678(self, **kw):
        return (202, {}, None)

    def delete_servers_1234_metadata_key1(self, **kw):
        return (204, {}, None)

    def delete_servers_1234_metadata_key2(self, **kw):
        return (204, {}, None)

    def post_servers_1234_metadata(self, **kw):
        return (204, {}, {'metadata': {'test_key': 'test_value'}})

    def get_servers_1234_diagnostics(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def post_servers_uuid1_metadata(self, **kw):
        return (204, {}, {'metadata': {'key1': 'val1'}})

    def post_servers_uuid2_metadata(self, **kw):
        return (204, {}, {'metadata': {'key1': 'val1'}})

    def post_servers_uuid3_metadata(self, **kw):
        return (204, {}, {'metadata': {'key1': 'val1'}})

    def post_servers_uuid4_metadata(self, **kw):
        return (204, {}, {'metadata': {'key1': 'val1'}})

    def post_servers_uuid5_metadata(self, **kw):
        return (204, {}, {'metadata': {'key1': 'val1'}})

    def post_servers_uuid6_metadata(self, **kw):
        return (204, {}, {'metadata': {'key1': 'val1'}})

    def delete_servers_uuid1_metadata_key1(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def delete_servers_uuid2_metadata_key1(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def delete_servers_uuid3_metadata_key1(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def delete_servers_uuid4_metadata_key1(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def delete_servers_uuid5_metadata_key1(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def delete_servers_uuid6_metadata_key1(self, **kw):
        return (200, {}, {'data': 'Fake diagnostics'})

    def get_servers_1234_os_security_groups(self, **kw):
        return (200, {}, {
            "security_groups": [{
                'id': 1,
                'name': 'securitygroup1',
                'description': 'FAKE_SECURITY_GROUP',
                'tenant_id': '4ffc664c198e435e9853f2538fbcd7a7',
                'rules': []}]
        })

    def get_servers_1234_topology(self, **kw):
        return 200, {}, SERVER_TOPOLOGY

    #
    # Server password
    #

    # Testing with the following password and key
    #
    # Clear password: FooBar123
    #
    # RSA Private Key: novaclient/tests/unit/idfake.pem
    #
    # Encrypted password
    # OIuEuQttO8Rk93BcKlwHQsziDAnkAm/V6V8VPToA8ZeUaUBWwS0gwo2K6Y61Z96r
    # qG447iRz0uTEEYq3RAYJk1mh3mMIRVl27t8MtIecR5ggVVbz1S9AwXJQypDKl0ho
    # QFvhCBcMWPohyGewDJOhDbtuN1IoFI9G55ZvFwCm5y7m7B2aVcoLeIsJZE4PLsIw
    # /y5a6Z3/AoJZYGG7IH5WN88UROU3B9JZGFB2qtPLQTOvDMZLUhoPRIJeHiVSlo1N
    # tI2/++UsXVg3ow6ItqCJGgdNuGG5JB+bslDHWPxROpesEIHdczk46HCpHQN8f1sk
    # Hi/fmZZNQQqj1Ijq0caOIw==
    def get_servers_1234_os_server_password(self, **kw):
        return (200, {}, {
            'password':
            'OIuEuQttO8Rk93BcKlwHQsziDAnkAm/V6V8VPToA8ZeUaUBWwS0gwo2K6Y61Z96r'
            'qG447iRz0uTEEYq3RAYJk1mh3mMIRVl27t8MtIecR5ggVVbz1S9AwXJQypDKl0ho'
            'QFvhCBcMWPohyGewDJOhDbtuN1IoFI9G55ZvFwCm5y7m7B2aVcoLeIsJZE4PLsIw'
            '/y5a6Z3/AoJZYGG7IH5WN88UROU3B9JZGFB2qtPLQTOvDMZLUhoPRIJeHiVSlo1N'
            'tI2/++UsXVg3ow6ItqCJGgdNuGG5JB+bslDHWPxROpesEIHdczk46HCpHQN8f1sk'
            'Hi/fmZZNQQqj1Ijq0caOIw=='})

    def delete_servers_1234_os_server_password(self, **kw):
        return (202, {}, None)

    #
    # Server actions
    #

    none_actions = ['revertResize', 'os-stop', 'os-start',
                    'forceDelete', 'restore', 'pause', 'unpause', 'unlock',
                    'unrescue', 'resume', 'suspend', 'shelve',
                    'shelveOffload', 'resetNetwork']
    type_actions = ['os-getVNCConsole', 'os-getSPICEConsole',
                    'os-getRDPConsole']

    @classmethod
    def check_server_actions(cls, body):
        action = list(body)[0]
        if action == 'reboot':
            assert list(body[action]) == ['type']
            assert body[action]['type'] in ['HARD', 'SOFT']
        elif action == 'resize':
            assert 'flavorRef' in body[action]
        elif action in cls.none_actions:
            assert body[action] is None
        elif action == 'changePassword':
            assert list(body[action]) == ['adminPass']
        elif action in cls.type_actions:
            assert list(body[action]) == ['type']
        elif action == 'os-resetState':
            assert list(body[action]) == ['state']
        elif action == 'resetNetwork':
            assert body[action] is None
        elif action in ['addSecurityGroup', 'removeSecurityGroup']:
            assert list(body[action]) == ['name']
        elif action == 'trigger_crash_dump':
            assert body[action] is None
        else:
            return False
        return True

    def post_servers_1234_action(self, body, **kw):
        _headers = dict()
        _body = None
        resp = 202
        assert len(body.keys()) == 1
        action = list(body)[0]

        if self.check_server_actions(body):
            # NOTE(snikitin): No need to do any operations here. This 'pass'
            # is needed to avoid AssertionError in the last 'else' statement
            # if we found 'action' in method check_server_actions and
            # raise AssertionError if we didn't find 'action' at all.
            pass
        elif action == 'os-migrateLive':
            expected = set(['host', 'block_migration'])
            if self.api_version >= api_versions.APIVersion("2.30"):
                if 'force' in body[action].keys():
                    # force can be optional
                    expected.add('force')
            if self.api_version < api_versions.APIVersion("2.25"):
                expected.add('disk_over_commit')
            assert set(body[action].keys()) == expected
        elif action == 'migrate':
            if self.api_version < api_versions.APIVersion("2.56"):
                assert body[action] is None
            else:
                expected = set()
                if 'host' in body[action].keys():
                    # host can be optional
                    expected.add('host')
                assert set(body[action].keys()) == expected
        elif action == 'lock':
            if self.api_version < api_versions.APIVersion("2.73"):
                assert body[action] is None
            else:
                # In 2.73 and above, we allow body to be one of these:
                # a) {'lock': None}
                # b) {'lock': {}}
                # c) {'lock': {locked_reason': 'blah'}}
                if body[action] is not None:
                    expected = set()
                    if 'locked_reason' in body[action].keys():
                        # reason can be optional
                        expected.add('locked_reason')
                    assert set(body[action].keys()) == expected
                else:
                    assert body[action] is None
        elif action == 'unshelve':
            if self.api_version < api_versions.APIVersion("2.77"):
                assert body[action] is None
            else:
                # In 2.77 to 2.91, we allow body to be one of these:
                # {'unshelve': None}
                # {'unshelve': {'availability_zone': 'foo-az'}}
                if body[action] is not None:
                    assert set(body[action].keys()) == set(
                        ['availability_zone'])
        elif action == 'rebuild':
            body = body[action]
            adminPass = body.get('adminPass', 'randompassword')
            assert 'imageRef' in body
            _body = self.get_servers_1234()[2]
            _body['server']['adminPass'] = adminPass
        elif action == 'confirmResize':
            assert body[action] is None
            # This one method returns a different response code
            return (204, {}, None)
        elif action == 'rescue':
            if body[action]:
                keys = set(body[action].keys())
                assert not (keys - set(['adminPass', 'rescue_image_ref']))
            else:
                assert body[action] is None
            _body = {'adminPass': 'RescuePassword'}
        elif action == 'createImage':
            assert set(body[action].keys()) == set(['name', 'metadata'])
            if self.api_version < api_versions.APIVersion('2.45'):
                _headers = dict(location="http://blah/images/%s" %
                                FAKE_IMAGE_UUID_SNAPSHOT)
            else:
                _body = {'image_id': FAKE_IMAGE_UUID_SNAPSHOT}
            if body[action]['name'] == 'mysnapshot_deleted':
                _headers = dict(location="http://blah/images/%s" %
                                FAKE_IMAGE_UUID_SNAP_DEL)
        elif action == 'createBackup':
            assert set(body[action].keys()) == set(['name', 'backup_type',
                                                    'rotation'])
            if self.api_version < api_versions.APIVersion('2.45'):
                _headers = dict(location="http://blah/images/%s" %
                                FAKE_IMAGE_UUID_BACKUP)
            else:
                _body = {'image_id': FAKE_IMAGE_UUID_BACKUP}
        elif action == 'os-getConsoleOutput':
            assert list(body[action]) == ['length']
            return (202, {}, {'output': 'foo'})
        elif action == 'evacuate':
            keys = list(body[action])
            if 'adminPass' in keys:
                keys.remove('adminPass')
            if 'host' in keys:
                keys.remove('host')
            if 'onSharedStorage' in keys:
                keys.remove('onSharedStorage')
            if 'force' in keys:
                keys.remove('force')
            assert set(keys) == set()
        else:
            raise AssertionError("Unexpected server action: %s" % action)
        _headers.update(FAKE_RESPONSE_HEADERS)
        return (resp, _headers, _body)

    def post_servers_5678_action(self, body, **kw):
        return self.post_servers_1234_action(body, **kw)

    #
    # Flavors
    #

    def get_flavors(self, **kw):
        status, header, flavors = self.get_flavors_detail(**kw)
        included_fields = ['id', 'name']
        if self.api_version >= api_versions.APIVersion('2.55'):
            included_fields.append('description')
        if self.api_version >= api_versions.APIVersion('2.61'):
            included_fields.append('extra_specs')
        for flavor in flavors['flavors']:
            for k in list(flavor):
                if k not in included_fields:
                    del flavor[k]

        return (200, FAKE_RESPONSE_HEADERS, flavors)

    def get_flavors_detail(self, **kw):
        flavors = {'flavors': [
            {'id': 1, 'name': '256 MiB Server', 'ram': 256, 'disk': 10,
             'OS-FLV-EXT-DATA:ephemeral': 10,
             'os-flavor-access:is_public': True,
             'links': {}},
            {'id': 2, 'name': '512 MiB Server', 'ram': 512, 'disk': 20,
             'OS-FLV-EXT-DATA:ephemeral': 20,
             'os-flavor-access:is_public': False,
             'links': {}},
            {'id': 4, 'name': '1024 MiB Server', 'ram': 1024, 'disk': 10,
             'OS-FLV-EXT-DATA:ephemeral': 10,
             'os-flavor-access:is_public': True,
             'links': {}},
            {'id': 'aa1', 'name': '128 MiB Server', 'ram': 128, 'disk': 0,
             'OS-FLV-EXT-DATA:ephemeral': 0,
             'os-flavor-access:is_public': True,
             'links': {}}
        ]}

        if 'is_public' not in kw:
            filter_is_public = True
        else:
            if kw['is_public'].lower() == 'none':
                filter_is_public = None
            else:
                filter_is_public = strutils.bool_from_string(kw['is_public'],
                                                             True)

        if filter_is_public is not None:
            if filter_is_public:
                flavors['flavors'] = [
                    v for v in flavors['flavors']
                    if v['os-flavor-access:is_public']
                ]
            else:
                flavors['flavors'] = [
                    v for v in flavors['flavors']
                    if not v['os-flavor-access:is_public']
                ]

        # Add description in the response for all flavors.
        if self.api_version >= api_versions.APIVersion('2.55'):
            for flavor in flavors['flavors']:
                flavor['description'] = None
            # Add a new flavor that is a copy of the first but with a different
            # name, flavorid and a description set.
            new_flavor = copy.deepcopy(flavors['flavors'][0])
            new_flavor['id'] = 'with-description'
            new_flavor['name'] = 'with-description'
            new_flavor['description'] = 'test description'
            flavors['flavors'].append(new_flavor)

        # Add extra_specs in the response for all flavors.
        if self.api_version >= api_versions.APIVersion('2.61'):
            for flavor in flavors['flavors']:
                flavor['extra_specs'] = {'test': 'value'}

        return (200, FAKE_RESPONSE_HEADERS, flavors)

    def get_flavors_1(self, **kw):
        return (
            200,
            FAKE_RESPONSE_HEADERS,
            {'flavor':
                self.get_flavors_detail(is_public='None')[2]['flavors'][0]}
        )

    def get_flavors_2(self, **kw):
        return (
            200,
            {},
            {'flavor':
                self.get_flavors_detail(is_public='None')[2]['flavors'][1]}
        )

    def get_flavors_3(self, **kw):
        # Diablo has no ephemeral
        return (
            200,
            FAKE_RESPONSE_HEADERS,
            {'flavor': {
                'id': 3,
                'name': '256 MiB Server',
                'ram': 256,
                'disk': 10,
            }},
        )

    def get_flavors_512_MiB_Server(self, **kw):
        raise exceptions.NotFound('404')

    def get_flavors_128_MiB_Server(self, **kw):
        raise exceptions.NotFound('404')

    def get_flavors_80645cf4_6ad3_410a_bbc8_6f3e1e291f51(self, **kw):
        raise exceptions.NotFound('404')

    def get_flavors_aa1(self, **kw):
        # Alphanumeric flavor id are allowed.
        return (
            200,
            FAKE_RESPONSE_HEADERS,
            {'flavor':
                self.get_flavors_detail(is_public='None')[2]['flavors'][3]}
        )

    def get_flavors_4(self, **kw):
        return (
            200,
            {},
            {'flavor':
                self.get_flavors_detail(is_public='None')[2]['flavors'][2]}
        )

    def get_flavors_with_description(self, **kw):
        return (
            200,
            {},
            {'flavor':
                self.get_flavors_detail(is_public='None')[2]['flavors'][-1]}
        )

    def delete_flavors_flavordelete(self, **kw):
        return (202, FAKE_RESPONSE_HEADERS, None)

    def delete_flavors_2(self, **kw):
        return (202, FAKE_RESPONSE_HEADERS, None)

    def post_flavors(self, body, **kw):
        return (
            202,
            FAKE_RESPONSE_HEADERS,
            {'flavor':
                self.get_flavors_detail(is_public='None')[2]['flavors'][0]}
        )

    def put_flavors_with_description(self, body, **kw):
        assert 'flavor' in body
        assert 'description' in body['flavor']
        flavor = self.get_flavors_with_description(**kw)[2]
        # Fake out the actual update of the flavor description for the response
        flavor['description'] = body['flavor']['description']
        return (200, {}, {'flavor': flavor})

    def get_flavors_1_os_extra_specs(self, **kw):
        return (
            200,
            {},
            {'extra_specs': {"k1": "v1"}})

    def get_flavors_2_os_extra_specs(self, **kw):
        return (
            200,
            {},
            {'extra_specs': {"k2": "v2"}})

    def get_flavors_aa1_os_extra_specs(self, **kw):
        return (
            200, {},
            {'extra_specs': {"k3": "v3"}})

    def get_flavors_4_os_extra_specs(self, **kw):
        return (
            200,
            {},
            {'extra_specs': {"k4": "v4"}})

    def post_flavors_1_os_extra_specs(self, body, **kw):
        assert list(body) == ['extra_specs']
        fakes.assert_has_keys(body['extra_specs'],
                              required=['k1'])
        return (
            200,
            FAKE_RESPONSE_HEADERS,
            {'extra_specs': {"k1": "v1"}})

    def post_flavors_4_os_extra_specs(self, body, **kw):
        assert list(body) == ['extra_specs']

        return (
            200,
            FAKE_RESPONSE_HEADERS,
            body)

    def delete_flavors_1_os_extra_specs_k1(self, **kw):
        return (204, {}, None)

    #
    # Flavor access
    #

    def get_flavors_2_os_flavor_access(self, **kw):
        return (
            200, FAKE_RESPONSE_HEADERS,
            {'flavor_access': [{'flavor_id': '2', 'tenant_id': 'proj1'},
                               {'flavor_id': '2', 'tenant_id': 'proj2'}]})

    def post_flavors_2_action(self, body, **kw):
        return (202, FAKE_RESPONSE_HEADERS,
                self.get_flavors_2_os_flavor_access()[2])

    #
    # Images
    #
    def get_images(self, **kw):
        images = [
            {
                "id": FAKE_IMAGE_UUID_SNAPSHOT,
                "name": "My Server Backup",
                "serverId": '1234',
                "updated": "2010-10-10T12:00:00Z",
                "created": "2010-08-10T12:00:00Z",
                "status": "SAVING",
                "progress": 80,
                "links": {},
            },
            {
                "id": FAKE_IMAGE_UUID_SNAP_DEL,
                "name": "My Server Backup Deleted",
                "serverId": '1234',
                "updated": "2010-10-10T12:00:00Z",
                "created": "2010-08-10T12:00:00Z",
                "status": "DELETED",
                "fault": {'message': 'Image has been deleted.'},
                "links": {},
            },
            {
                'id': FAKE_IMAGE_UUID_1,
                'name': 'CentOS 5.2',
                "updated": "2010-10-10T12:00:00Z",
                "created": "2010-08-10T12:00:00Z",
                "status": "ACTIVE",
                "test_key": "test_value",
                "links": {},
            },
            {
                "id": FAKE_IMAGE_UUID_2,
                "name": "My Server Backup",
                "serverId": '1234',
                "updated": "2010-10-10T12:00:00Z",
                "created": "2010-08-10T12:00:00Z",
                "status": "SAVING",
                "progress": 80,
                "links": {},
            },
            {
                "id": FAKE_IMAGE_UUID_BACKUP,
                "name": "back1",
                "serverId": '1234',
                "updated": "2010-10-10T12:00:00Z",
                "created": "2010-08-10T12:00:00Z",
                "status": "SAVING",
                "progress": 80,
                "links": {},
            },
        ]

        if 'id' in kw:
            requested = kw['id'].replace('in:', '').split(',')
            images = [i for i in images if i['id'] in requested]
        if 'names' in kw:
            requested = kw['names'].replace('in:', '').split(',')
            images = [i for i in images if i['name'] in requested]

        return (200, {}, {'images': images})

    def get_images_555cae93_fb41_4145_9c52_f5b923538a26(self, **kw):
        return (200, {}, {'image': self.get_images()[2]['images'][0]})

    def get_images_55bb23af_97a4_4068_bdf8_f10c62880ddf(self, **kw):
        return (200, {}, {'image': self.get_images()[2]['images'][1]})

    def get_images_c99d7632_bd66_4be9_aed5_3dd14b223a76(self, **kw):
        return (200, {}, {'image': self.get_images()[2]['images'][2]})

    def get_images_f27f479a_ddda_419a_9bbc_d6b56b210161(self, **kw):
        return (200, {}, {'image': self.get_images()[2]['images'][3]})

    def get_images_2f87e889_41a4_4778_8553_83f5eea68c5d(self, **kw):
        return (200, {}, {'image': self.get_images()[2]['images'][4]})

    def get_images_3e861307_73a6_4d1f_8d68_f68b03223032(self):
        raise exceptions.NotFound('404')

    #
    # Keypairs
    #
    def get_os_keypairs_test(self, *kw):
        return (200, {}, {'keypair':
                          self.get_os_keypairs()[2]['keypairs'][0]['keypair']})

    def get_os_keypairs(self, user_id=None, limit=None, marker=None, *kw):
        return (200, {}, {
            "keypairs": [{"keypair": {
                "public_key": "FAKE_SSH_RSA",
                "private_key": "FAKE_PRIVATE_KEY",
                "user_id": "81e373b596d6466e99c4896826abaa46",
                "name": "test",
                "deleted": False,
                "created_at": "2014-04-19T02:16:44.000000",
                "updated_at": "2014-04-19T10:12:3.000000",
                "fingerprint": "FAKE_KEYPAIR",
                "deleted_at": None,
                "id": 4}}
            ]})

    def delete_os_keypairs_test(self, **kw):
        return (202, {}, None)

    def post_os_keypairs(self, body, **kw):
        assert list(body) == ['keypair']
        fakes.assert_has_keys(body['keypair'],
                              required=['name'])
        r = {'keypair': self.get_os_keypairs()[2]['keypairs'][0]['keypair']}
        return (202, {}, r)

    #
    # Quotas
    #

    def get_os_quota_sets_tenant_id(self, **kw):
        return (200, {}, {
            'quota_set': {
                'tenant_id': 'test',
                'metadata_items': [],
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'keypairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def get_os_quota_sets_97f4c221bff44578b0300df4ef119353(self, **kw):
        return (200, {}, {
            'quota_set': {
                'tenant_id': '97f4c221bff44578b0300df4ef119353',
                'metadata_items': [],
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'keypairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def get_os_quota_sets_97f4c221bff44578b0300df4ef119353_detail(self, **kw):
        return (200, {}, {
            'quota_set': {
                'tenant_id': '97f4c221bff44578b0300df4ef119353',
                'cores': {
                    'in_use': 0,
                    'limit': 20,
                    'reserved': 0
                },
                'fixed_ips': {
                    'in_use': 0,
                    'limit': -1,
                    'reserved': 0
                },
                'floating_ips': {
                    'in_use': 0,
                    'limit': 10,
                    'reserved': 0
                },
                'injected_file_content_bytes': {
                    'in_use': 0,
                    'limit': 10240,
                    'reserved': 0
                },
                'injected_file_path_bytes': {
                    'in_use': 0,
                    'limit': 255,
                    'reserved': 0
                },
                'injected_files': {
                    'in_use': 0,
                    'limit': 5,
                    'reserved': 0
                },
                'instances': {
                    'in_use': 0,
                    'limit': 10,
                    'reserved': 0
                },
                'key_pairs': {
                    'in_use': 0,
                    'limit': 100,
                    'reserved': 0
                },
                'metadata_items': {
                    'in_use': 0,
                    'limit': 128,
                    'reserved': 0
                },
                'ram': {
                    'in_use': 0,
                    'limit': 51200,
                    'reserved': 0
                },
                'security_group_rules': {
                    'in_use': 0,
                    'limit': 20,
                    'reserved': 0
                },
                'security_groups': {
                    'in_use': 0,
                    'limit': 10,
                    'reserved': 0
                },
                'server_group_members': {
                    'in_use': 0,
                    'limit': 10,
                    'reserved': 0
                },
                'server_groups': {
                    'in_use': 0,
                    'limit': 10,
                    'reserved': 0
                }
            }})

    def get_os_quota_sets_97f4c221bff44578b0300df4ef119353_defaults(self):
        return (200, {}, {
            'quota_set': {
                'tenant_id': 'test',
                'metadata_items': [],
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'keypairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def get_os_quota_sets_tenant_id_defaults(self):
        return (200, {}, {
            'quota_set': {
                'tenant_id': 'test',
                'metadata_items': [],
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'keypairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def put_os_quota_sets_97f4c221bff44578b0300df4ef119353(self, body, **kw):
        assert list(body) == ['quota_set']
        fakes.assert_has_keys(body['quota_set'])
        return (200, {}, {
            'quota_set': {
                'tenant_id': '97f4c221bff44578b0300df4ef119353',
                'metadata_items': [],
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'keypairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def delete_os_quota_sets_97f4c221bff44578b0300df4ef119353(self, **kw):
        return (202, {}, {})

    #
    # Quota Classes
    #

    def get_os_quota_class_sets_test(self, **kw):
        # 2.57 removes injected_file* entries from the response.
        if self.api_version >= api_versions.APIVersion('2.57'):
            return (200, FAKE_RESPONSE_HEADERS, {
                'quota_class_set': {
                    'id': 'test',
                    'metadata_items': 1,
                    'ram': 1,
                    'instances': 1,
                    'cores': 1,
                    'key_pairs': 1,
                    'server_groups': 1,
                    'server_group_members': 1}})

        if self.api_version >= api_versions.APIVersion('2.50'):
            return (200, FAKE_RESPONSE_HEADERS, {
                'quota_class_set': {
                    'id': 'test',
                    'metadata_items': 1,
                    'injected_file_content_bytes': 1,
                    'injected_file_path_bytes': 1,
                    'ram': 1,
                    'instances': 1,
                    'injected_files': 1,
                    'cores': 1,
                    'key_pairs': 1,
                    'server_groups': 1,
                    'server_group_members': 1}})
        return (200, FAKE_RESPONSE_HEADERS, {
            'quota_class_set': {
                'id': 'test',
                'metadata_items': 1,
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'fixed_ips': -1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'key_pairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def put_os_quota_class_sets_test(self, body, **kw):
        assert list(body) == ['quota_class_set']
        # 2.57 removes injected_file* entries from the response.
        if self.api_version >= api_versions.APIVersion('2.57'):
            return (200, {}, {
                'quota_class_set': {
                    'metadata_items': 1,
                    'ram': 1,
                    'instances': 1,
                    'cores': 1,
                    'key_pairs': 1,
                    'server_groups': 1,
                    'server_group_members': 1}})

        if self.api_version >= api_versions.APIVersion('2.50'):
            return (200, {}, {
                'quota_class_set': {
                    'metadata_items': 1,
                    'injected_file_content_bytes': 1,
                    'injected_file_path_bytes': 1,
                    'ram': 1,
                    'instances': 1,
                    'injected_files': 1,
                    'cores': 1,
                    'key_pairs': 1,
                    'server_groups': 1,
                    'server_group_members': 1}})
        return (200, {}, {
            'quota_class_set': {
                'metadata_items': 1,
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'fixed_ips': -1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'key_pairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    def put_os_quota_class_sets_97f4c221bff44578b0300df4ef119353(self,
                                                                 body, **kw):
        assert list(body) == ['quota_class_set']
        return (200, {}, {
            'quota_class_set': {
                'metadata_items': 1,
                'injected_file_content_bytes': 1,
                'injected_file_path_bytes': 1,
                'ram': 1,
                'floating_ips': 1,
                'instances': 1,
                'injected_files': 1,
                'cores': 1,
                'key_pairs': 1,
                'security_groups': 1,
                'security_group_rules': 1}})

    #
    # Tenant Usage
    #
    def get_os_simple_tenant_usage(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {'tenant_usages': [{
            'total_memory_mb_usage': 25451.762807466665,
            'total_vcpus_usage': 49.71047423333333,
            'total_hours': 49.71047423333333,
            'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
            'stop': '2012-01-22 19:48:41.750722',
            'server_usages': [{
                'hours': 49.71047423333333,
                'uptime': 27035,
                'local_gb': 0,
                'ended_at': None,
                'name': 'f15image1',
                'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
                'instance_id': 'f079e394-1111-457b-b350-bb5ecc685cdd',
                'vcpus': 1,
                'memory_mb': 512,
                'state': 'active',
                'flavor': 'm1.tiny',
                'started_at': '2012-01-20 18:06:06.479998',
            }],
            'start': '2011-12-25 19:48:41.750687',
            'total_local_gb_usage': 0.0}]})

    def get_os_simple_tenant_usage_next(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {'tenant_usages': [{
            'total_memory_mb_usage': 25451.762807466665,
            'total_vcpus_usage': 49.71047423333333,
            'total_hours': 49.71047423333333,
            'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
            'stop': '2012-01-22 19:48:41.750722',
            'server_usages': [{
                'hours': 49.71047423333333,
                'uptime': 27035,
                'local_gb': 0,
                'ended_at': None,
                'name': 'f15image1',
                'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
                'instance_id': 'f079e394-2222-457b-b350-bb5ecc685cdd',
                'vcpus': 1,
                'memory_mb': 512,
                'state': 'active',
                'flavor': 'm1.tiny',
                'started_at': '2012-01-20 18:06:06.479998',
            }],
            'start': '2011-12-25 19:48:41.750687',
            'total_local_gb_usage': 0.0}]})

    def get_os_simple_tenant_usage_next_next(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {'tenant_usages': []})

    def get_os_simple_tenant_usage_tenantfoo(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {'tenant_usage': {
            'total_memory_mb_usage': 25451.762807466665,
            'total_vcpus_usage': 49.71047423333333,
            'total_hours': 49.71047423333333,
            'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
            'stop': '2012-01-22 19:48:41.750722',
            'server_usages': [{
                'hours': 49.71047423333333,
                'uptime': 27035, 'local_gb': 0,
                'ended_at': None,
                'name': 'f15image1',
                'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
                'instance_id': 'f079e394-1111-457b-b350-bb5ecc685cdd',
                'vcpus': 1, 'memory_mb': 512,
                'state': 'active',
                'flavor': 'm1.tiny',
                'started_at': '2012-01-20 18:06:06.479998',
            }],
            'start': '2011-12-25 19:48:41.750687',
            'total_local_gb_usage': 0.0}})

    def get_os_simple_tenant_usage_test(self, **kw):
        return (200, {}, {'tenant_usage': {
            'total_memory_mb_usage': 25451.762807466665,
            'total_vcpus_usage': 49.71047423333333,
            'total_hours': 49.71047423333333,
            'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
            'stop': '2012-01-22 19:48:41.750722',
            'server_usages': [{
                'hours': 49.71047423333333,
                'uptime': 27035, 'local_gb': 0,
                'ended_at': None,
                'name': 'f15image1',
                'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
                'instance_id': 'f079e394-1111-457b-b350-bb5ecc685cdd',
                'vcpus': 1, 'memory_mb': 512,
                'state': 'active',
                'flavor': 'm1.tiny',
                'started_at': '2012-01-20 18:06:06.479998',
            }],
            'start': '2011-12-25 19:48:41.750687',
            'total_local_gb_usage': 0.0}})

    def get_os_simple_tenant_usage_tenant_id(self, **kw):
        return (200, {}, {'tenant_usage': {
            'total_memory_mb_usage': 25451.762807466665,
            'total_vcpus_usage': 49.71047423333333,
            'total_hours': 49.71047423333333,
            'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
            'stop': '2012-01-22 19:48:41.750722',
            'server_usages': [{
                'hours': 49.71047423333333,
                'uptime': 27035, 'local_gb': 0,
                'ended_at': None,
                'name': 'f15image1',
                'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
                'instance_id': 'f079e394-1111-457b-b350-bb5ecc685cdd',
                'vcpus': 1, 'memory_mb': 512,
                'state': 'active',
                'flavor': 'm1.tiny',
                'started_at': '2012-01-20 18:06:06.479998',
            }],
            'start': '2011-12-25 19:48:41.750687',
            'total_local_gb_usage': 0.0}})

    def get_os_simple_tenant_usage_tenant_id_next(self, **kw):
        return (200, {}, {'tenant_usage': {
            'total_memory_mb_usage': 25451.762807466665,
            'total_vcpus_usage': 49.71047423333333,
            'total_hours': 49.71047423333333,
            'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
            'stop': '2012-01-22 19:48:41.750722',
            'server_usages': [{
                'hours': 49.71047423333333,
                'uptime': 27035, 'local_gb': 0,
                'ended_at': None,
                'name': 'f15image1',
                'tenant_id': '7b0a1d73f8fb41718f3343c207597869',
                'instance_id': 'f079e394-2222-457b-b350-bb5ecc685cdd',
                'vcpus': 1, 'memory_mb': 512,
                'state': 'active',
                'flavor': 'm1.tiny',
                'started_at': '2012-01-20 18:06:06.479998',
            }],
            'start': '2011-12-25 19:48:41.750687',
            'total_local_gb_usage': 0.0}})

    def get_os_simple_tenant_usage_tenant_id_next_next(self, **kw):
        return (200, {}, {'tenant_usage': {}})

    #
    # Aggregates
    #

    def get_os_aggregates(self, *kw):
        response = (200, {}, {"aggregates": [
            {'id': '1',
             'name': 'test',
             'availability_zone': 'nova1'},
            {'id': '2',
             'name': 'test2',
             'availability_zone': 'nova1'},
            {'id': '3',
             'name': 'test3',
             'metadata': {'test': "dup", "none_key": "Nine"}},
        ]})
        # microversion >= 2.41 returns the uuid in the response
        if self.api_version >= api_versions.APIVersion('2.41'):
            aggregates = response[2]['aggregates']
            aggregates[0]['uuid'] = '80785864-087b-45a5-a433-b20eac9b58aa'
            aggregates[1]['uuid'] = '30827713-5957-4b68-8fd3-ccaddb568c24'
            aggregates[2]['uuid'] = '9a651b22-ce3f-4a87-acd7-98446ef591c4'
        return response

    def _return_aggregate(self):
        r = {'aggregate': self.get_os_aggregates()[2]['aggregates'][0]}
        return (200, {}, r)

    def _return_aggregate_3(self):
        r = {'aggregate': self.get_os_aggregates()[2]['aggregates'][2]}
        return (200, {}, r)

    def get_os_aggregates_1(self, **kw):
        return self._return_aggregate()

    def get_os_aggregates_3(self, **kw):
        return self._return_aggregate_3()

    def post_os_aggregates(self, body, **kw):
        return self._return_aggregate()

    def put_os_aggregates_1(self, body, **kw):
        return self._return_aggregate()

    def post_os_aggregates_1_action(self, body, **kw):
        return self._return_aggregate()

    def post_os_aggregates_3_action(self, body, **kw):
        return self._return_aggregate_3()

    def delete_os_aggregates_1(self, **kw):
        return (202, {}, None)

    def post_os_aggregates_1_images(self, body, **kw):
        return (202, {}, None)

    #
    # Services
    #
    def get_os_services(self, **kw):
        host = kw.get('host', 'host1')
        binary = kw.get('binary', 'nova-compute')
        if self.api_version >= api_versions.APIVersion('2.53'):
            service_id_1 = FAKE_SERVICE_UUID_1
            service_id_2 = FAKE_SERVICE_UUID_2
        else:
            service_id_1 = 1
            service_id_2 = 2
        services = {
            'services': [
                {'binary': binary,
                 'host': host,
                 'zone': 'nova',
                 'status': 'enabled',
                 'state': 'up',
                 'updated_at': datetime.datetime(
                     2012, 10, 29, 13, 42, 2),
                 'id': service_id_1},
                {'binary': binary,
                 'host': host,
                 'zone': 'nova',
                 'status': 'disabled',
                 'state': 'down',
                 'updated_at': datetime.datetime(
                     2012, 9, 18, 8, 3, 38),
                 'id': service_id_2},
            ]
        }
        if self.api_version >= api_versions.APIVersion('2.69'):
            services['services'].append(
                {
                    "binary": "nova-compute",
                    "host": "host-down",
                    "status": "UNKNOWN"
                }
            )
        return (200, FAKE_RESPONSE_HEADERS, services)

    def put_os_services_enable(self, body, **kw):
        return (200, FAKE_RESPONSE_HEADERS,
                {'service': {'host': body['host'],
                             'binary': body['binary'],
                             'status': 'enabled'}})

    def put_os_services_disable(self, body, **kw):
        return (200, FAKE_RESPONSE_HEADERS,
                {'service': {'host': body['host'],
                             'binary': body['binary'],
                             'status': 'disabled'}})

    def put_os_services_disable_log_reason(self, body, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {'service': {
            'host': body['host'],
            'binary': body['binary'],
            'status': 'disabled',
            'disabled_reason': body['disabled_reason']}})

    def put_os_services_75e9eabc_ed3b_4f11_8bba_add1e7e7e2de(
            self, body, **kw):
        """This should only be called with microversion >= 2.53."""
        return (200, FAKE_RESPONSE_HEADERS, {'service': {
            'host': 'host1',
            'binary': 'nova-compute',
            'status': body.get('status', 'enabled'),
            'disabled_reason': body.get('disabled_reason'),
            'forced_down': body.get('forced_down', False)}})

    def delete_os_services_1(self, **kw):
        return (204, FAKE_RESPONSE_HEADERS, None)

    def delete_os_services_75e9eabc_ed3b_4f11_8bba_add1e7e7e2de(self, **kwarg):
        return (204, FAKE_RESPONSE_HEADERS, None)

    def put_os_services_force_down(self, body, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {'service': {
            'host': body['host'],
            'binary': body['binary'],
            'forced_down': False}})

    #
    # Hypervisors
    #

    def get_os_hypervisors(self, **kw):
        return (200, {}, {
            "hypervisors": [
                {'id': 1234, 'hypervisor_hostname': 'hyper1'},
                {'id': 5678, 'hypervisor_hostname': 'hyper2'}]})

    def get_os_hypervisors_statistics(self, **kw):
        return (200, {}, {
            "hypervisor_statistics": {
                'count': 2,
                'vcpus': 8,
                'memory_mb': 20 * 1024,
                'local_gb': 500,
                'vcpus_used': 4,
                'memory_mb_used': 10 * 1024,
                'local_gb_used': 250,
                'free_ram_mb': 10 * 1024,
                'free_disk_gb': 250,
                'current_workload': 4,
                'running_vms': 4,
                'disk_available_least': 200}
        })

    def get_os_hypervisors_hyper1(self, **kw):
        return (200, {}, {
            'hypervisor':
            {'id': 1234,
             'service': {'id': 1, 'host': 'compute1'},
             'vcpus': 4,
             'memory_mb': 10 * 1024,
             'local_gb': 250,
             'vcpus_used': 2,
             'memory_mb_used': 5 * 1024,
             'local_gb_used': 125,
             'hypervisor_type': "xen",
             'hypervisor_version': 3,
             'hypervisor_hostname': "hyper1",
             'free_ram_mb': 5 * 1024,
             'free_disk_gb': 125,
             'current_workload': 2,
             'running_vms': 2,
             'cpu_info': 'cpu_info',
             'disk_available_least': 100}})

    def get_os_hypervisors_region_child_1(self, **kw):
        return (200, {}, {
            'hypervisor':
            {'id': 'region!child@1',
             'service': {'id': 1, 'host': 'compute1'},
             'vcpus': 4,
             'memory_mb': 10 * 1024,
             'local_gb': 250,
             'vcpus_used': 2,
             'memory_mb_used': 5 * 1024,
             'local_gb_used': 125,
             'hypervisor_type': "xen",
             'hypervisor_version': 3,
             'hypervisor_hostname': "hyper1",
             'free_ram_mb': 5 * 1024,
             'free_disk_gb': 125,
             'current_workload': 2,
             'running_vms': 2,
             'cpu_info': 'cpu_info',
             'disk_available_least': 100}})

    def get_os_hypervisors_hyper_search(self, **kw):
        return (200, {}, {
            'hypervisors': [
                {'id': 1234, 'hypervisor_hostname': 'hyper1'},
                {'id': 5678, 'hypervisor_hostname': 'hyper2'}]})

    def get_os_hypervisors_hyper_servers(self, **kw):
        return (200, {}, {
            'hypervisors': [
                {'id': 1234,
                 'hypervisor_hostname': 'hyper1',
                 'servers': [
                     {'name': 'inst1', 'uuid': 'uuid1'},
                     {'name': 'inst2', 'uuid': 'uuid2'}]},
                {'id': 5678,
                 'hypervisor_hostname': 'hyper2',
                 'servers': [
                     {'name': 'inst3', 'uuid': 'uuid3'},
                     {'name': 'inst4', 'uuid': 'uuid4'}]}]
        })

    def get_os_hypervisors_hyper1_servers(self, **kw):
        return (200, {}, {
            'hypervisors': [
                {'id': 1234,
                 'hypervisor_hostname': 'hyper1',
                 'servers': [
                     {'name': 'inst1', 'uuid': 'uuid1'},
                     {'name': 'inst2', 'uuid': 'uuid2'}]}]
        })

    def get_os_hypervisors_hyper2_servers(self, **kw):
        return (200, {}, {
            'hypervisors': [
                {'id': 5678,
                 'hypervisor_hostname': 'hyper2',
                 'servers': [
                     {'name': 'inst3', 'uuid': 'uuid3'},
                     {'name': 'inst4', 'uuid': 'uuid4'}]}]
        })

    def get_os_hypervisors_hyper_no_servers_servers(self, **kw):
        return (200, {}, {'hypervisors':
                          [{'id': 1234, 'hypervisor_hostname': 'hyper1'}]})

    def get_os_hypervisors_1234(self, **kw):
        return (200, {}, {
            'hypervisor':
                {'id': 1234,
                 'service': {'id': 1, 'host': 'compute1'},
                 'vcpus': 4,
                 'memory_mb': 10 * 1024,
                 'local_gb': 250,
                 'vcpus_used': 2,
                 'memory_mb_used': 5 * 1024,
                 'local_gb_used': 125,
                 'hypervisor_type': "xen",
                 'hypervisor_version': 3,
                 'hypervisor_hostname': "hyper1",
                 'free_ram_mb': 5 * 1024,
                 'free_disk_gb': 125,
                 'current_workload': 2,
                 'running_vms': 2,
                 'cpu_info': 'cpu_info',
                 'disk_available_least': 100}})

    def get_os_hypervisors_1234_uptime(self, **kw):
        return (200, {}, {
            'hypervisor': {'id': 1234,
                           'hypervisor_hostname': "hyper1",
                           'uptime': "fake uptime"}})

    def get_os_hypervisors_region_child_1_uptime(self, **kw):
        return (200, {}, {
            'hypervisor': {'id': 'region!child@1',
                           'hypervisor_hostname': "hyper1",
                           'uptime': "fake uptime"}})

    def get_v2_0_networks(self, **kw):
        """Return neutron proxied networks.

        We establish a few different possible networks that we can get
        by name, which we can then call in tests. The only usage of
        this API should be for name -> id translation, however a full
        valid neutron block is provided for the private network to see
        the kinds of things that will be in that payload.
        """

        name = kw.get('name', "blank")

        networks_by_name = {
            'private': [
                {"status": "ACTIVE",
                 "router:external": False,
                 "availability_zone_hints": [],
                 "availability_zones": ["nova"],
                 "description": "",
                 "name": "private",
                 "subnets": ["64706c26-336c-4048-ab3c-23e3283bca2c",
                             "18512740-c760-4d5f-921f-668105c9ee44"],
                 "shared": False,
                 "tenant_id": "abd42f270bca43ea80fe4a166bc3536c",
                 "created_at": "2016-08-15T17:34:49",
                 "tags": [],
                 "ipv6_address_scope": None,
                 "updated_at": "2016-08-15T17:34:49",
                 "admin_state_up": True,
                 "ipv4_address_scope": None,
                 "port_security_enabled": True,
                 "mtu": 1450,
                 "id": "e43a56c7-11d4-45c9-8681-ddc8171b5850",
                 "revision": 2}],
            'duplicate': [
                {"status": "ACTIVE",
                 "id": "e43a56c7-11d4-45c9-8681-ddc8171b5850"},
                {"status": "ACTIVE",
                 "id": "f43a56c7-11d4-45c9-8681-ddc8171b5850"}],
            'blank': []
        }

        return (200, {}, {"networks": networks_by_name[name]})

    def get_os_availability_zone_detail(self, **kw):
        return (200, {}, {
            "availabilityZoneInfo": [
                {"zoneName": "zone-1",
                 "zoneState": {"available": True},
                 "hosts": {
                     "fake_host-1": {
                         "nova-compute": {
                             "active": True,
                             "available": True,
                             "updated_at": datetime.datetime(
                                 2012, 12, 26, 14, 45, 25, 0)}}}},
                {"zoneName": "internal",
                 "zoneState": {"available": True},
                 "hosts": {
                     "fake_host-1": {
                         "nova-sched": {
                             "active": True,
                             "available": True,
                             "updated_at": datetime.datetime(
                                 2012, 12, 26, 14, 45, 25, 0)}},
                     "fake_host-2": {
                         "nova-network": {
                             "active": True,
                             "available": False,
                             "updated_at": datetime.datetime(
                                 2012, 12, 26, 14, 45, 24, 0)}}}},
                {"zoneName": "zone-2",
                 "zoneState": {"available": False},
                 "hosts": None}]})

    def get_servers_1234_os_interface(self, **kw):
        attachments = {
            "interfaceAttachments": [
                {"port_state": "ACTIVE",
                 "net_id": "net-id-1",
                 "port_id": "port-id-1",
                 "mac_address": "aa:bb:cc:dd:ee:ff",
                 "fixed_ips": [{"ip_address": "1.2.3.4"}],
                 },
                {"port_state": "ACTIVE",
                 "net_id": "net-id-1",
                 "port_id": "port-id-1",
                 "mac_address": "aa:bb:cc:dd:ee:ff",
                 "fixed_ips": [{"ip_address": "1.2.3.4"}],
                 }
            ]
        }
        if self.api_version >= api_versions.APIVersion('2.70'):
            # Include the "tag" field in each attachment.
            for attachment in attachments['interfaceAttachments']:
                attachment['tag'] = 'test-tag'
        return (200, {}, attachments)

    def post_servers_1234_os_interface(self, **kw):
        attachment = {}
        if self.api_version >= api_versions.APIVersion('2.70'):
            # Include the "tag" field in the response.
            attachment['tag'] = 'test-tag'
        return (200, {}, {'interfaceAttachment': attachment})

    def delete_servers_1234_os_interface_port_id(self, **kw):
        return (200, {}, None)

    def post_servers_1234_os_volume_attachments(self, **kw):
        attachment = {"device": "/dev/vdb", "volumeId": 2}
        if self.api_version >= api_versions.APIVersion('2.70'):
            # Include the "tag" field in the response.
            attachment['tag'] = 'test-tag'

        if self.api_version >= api_versions.APIVersion('2.79'):
            # Include the "delete_on_termination" field in the
            # response.
            attachment['delete_on_termination'] = True
        return (200, FAKE_RESPONSE_HEADERS, {"volumeAttachment": attachment})

    def put_servers_1234_os_volume_attachments_Work(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS,
                {"volumeAttachment": {"volumeId": 2}})

    def get_servers_1234_os_volume_attachments(self, **kw):
        attachments = {
            "volumeAttachments": [
                {"display_name": "Work",
                 "display_description": "volume for work",
                 "status": "ATTACHED",
                 "id": "15e59938-07d5-11e1-90e3-e3dffe0c5983",
                 "created_at": "2011-09-09T00:00:00Z",
                 "attached": "2011-11-11T00:00:00Z",
                 "size": 1024,
                 "attachments": [{"id": "3333", "links": ''}],
                 "metadata": {}}
            ]
        }
        if self.api_version >= api_versions.APIVersion('2.70'):
            # Include the "tag" field in each attachment.
            for attachment in attachments['volumeAttachments']:
                attachment['tag'] = 'test-tag'

        if self.api_version >= api_versions.APIVersion('2.79'):
            # Include the "delete_on_termination" field in each
            # attachment.
            for attachment in attachments['volumeAttachments']:
                attachment['delete_on_termination'] = True
        return (200, FAKE_RESPONSE_HEADERS, attachments)

    def get_servers_1234_os_volume_attachments_Work(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {
            "volumeAttachment":
                {"display_name": "Work",
                 "display_description": "volume for work",
                 "status": "ATTACHED",
                 "id": "15e59938-07d5-11e1-90e3-e3dffe0c5983",
                 "created_at": "2011-09-09T00:00:00Z",
                 "attached": "2011-11-11T00:00:00Z",
                 "size": 1024,
                 "attachments": [{"id": "3333", "links": ''}],
                 "metadata": {}}})

    def delete_servers_1234_os_volume_attachments_Work(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {})

    def get_servers_1234_os_instance_actions(self, **kw):
        action = {"instance_uuid": "1234",
                  "user_id": "b968c25e04ab405f9fe4e6ca54cce9a5",
                  "start_time": "2013-03-25T13:45:09.000000",
                  "request_id": "req-abcde12345",
                  "action": "create",
                  "message": None,
                  "project_id": "04019601fe3648c0abd4f4abfb9e6106"}
        if self.api_version >= api_versions.APIVersion('2.58'):
            # This is intentionally different from the start_time.
            action['updated_at'] = '2013-03-25T13:50:09.000000'
        return (200, FAKE_RESPONSE_HEADERS, {
            "instanceActions": [action]})

    def get_servers_1234_os_instance_actions_req_abcde12345(self, **kw):
        action = {"instance_uuid": "1234",
                  "user_id": "b968c25e04ab405f9fe4e6ca54cce9a5",
                  "start_time": "2013-03-25T13:45:09.000000",
                  "request_id": "req-abcde12345",
                  "action": "create",
                  "message": None,
                  "project_id": "04019601fe3648c0abd4f4abfb9e6106"}
        if self.api_version >= api_versions.APIVersion('2.58'):
            action['updated_at'] = '2013-03-25T13:45:09.000000'
        return (200, FAKE_RESPONSE_HEADERS, {
            "instanceAction": action})

    def get_os_instance_usage_audit_log(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {
            "instance_usage_audit_logs": {
                "hosts_not_run": ["samplehost3"],
                "log": {
                    "samplehost0": {
                        "errors": 1,
                        "instances": 1,
                        "message": ("Instance usage audit ran for host "
                                    "samplehost0, 1 instances in 0.01 "
                                    "seconds."),
                        "state": "DONE"
                    },
                    "samplehost1": {
                        "errors": 1,
                        "instances": 2,
                        "message": ("Instance usage audit ran for host "
                                    "samplehost1, 2 instances in 0.01 "
                                    "seconds."),
                        "state": "DONE"
                    },
                    "samplehost2": {
                        "errors": 1,
                        "instances": 3,
                        "message": ("Instance usage audit ran for host "
                                    "samplehost2, 3 instances in 0.01 "
                                    "seconds."),
                        "state": "DONE"
                    },
                },
                "num_hosts": 4,
                "num_hosts_done": 3,
                "num_hosts_not_run": 1,
                "num_hosts_running": 0,
                "overall_status": "3 of 4 hosts done. 3 errors.",
                "period_beginning": "2012-06-01 00:00:00",
                "period_ending": "2012-07-01 00:00:00",
                "total_errors": 3,
                "total_instances": 6}})

    def get_os_instance_usage_audit_log_2016_12_10_13_59_59_999999(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {
            "instance_usage_audit_log": {
                "hosts_not_run": ["samplehost3"],
                "log": {
                    "samplehost0": {
                        "errors": 1,
                        "instances": 1,
                        "message": ("Instance usage audit ran for host "
                                    "samplehost0, 1 instances in 0.01 "
                                    "seconds."),
                        "state": "DONE"
                    },
                    "samplehost1": {
                        "errors": 1,
                        "instances": 2,
                        "message": ("Instance usage audit ran for host "
                                    "samplehost1, 2 instances in 0.01 "
                                    "seconds."),
                        "state": "DONE"
                    },
                    "samplehost2": {
                        "errors": 1,
                        "instances": 3,
                        "message": ("Instance usage audit ran for host "
                                    "samplehost2, 3 instances in 0.01 "
                                    "seconds."),
                        "state": "DONE"
                    },
                },
                "num_hosts": 4,
                "num_hosts_done": 3,
                "num_hosts_not_run": 1,
                "num_hosts_running": 0,
                "overall_status": "3 of 4 hosts done. 3 errors.",
                "period_beginning": "2012-06-01 00:00:00",
                "period_ending": "2012-07-01 00:00:00",
                "total_errors": 3,
                "total_instances": 6}})

    def get_os_instance_usage_audit_log__5Cu5de5_5Cu4f5c(self, **kw):
        return (400, {}, {"Invalid timestamp for date \u6f22\u5b57"})

    def post_servers_uuid1_action(self, **kw):
        return 202, {}, {}

    def post_servers_uuid2_action(self, **kw):
        return 202, {}, {}

    def post_servers_uuid3_action(self, **kw):
        return 202, {}, {}

    def post_servers_uuid4_action(self, **kw):
        return 202, {}, {}

    def post_servers_uuid5_action(self, **kw):
        return 202, {}, {}

    def post_servers_uuid6_action(self, **kw):
        return 202, {}, {}

    def get_os_migrations(self, **kw):
        migration1 = {
            "created_at": "2012-10-29T13:42:02.000000",
            "dest_compute": "compute2",
            "dest_host": "1.2.3.4",
            "dest_node": "node2",
            "id": '1234',
            "instance_uuid": "instance_id_123",
            "new_instance_type_id": 2,
            "old_instance_type_id": 1,
            "source_compute": "compute1",
            "source_node": "node1",
            "status": "Done",
            "updated_at": "2012-10-29T13:42:02.000000"
        }

        migration2 = {
            "created_at": "2012-10-29T13:42:02.000000",
            "dest_compute": "compute2",
            "dest_host": "1.2.3.4",
            "dest_node": "node2",
            "id": '1234',
            "instance_uuid": "instance_id_456",
            "new_instance_type_id": 2,
            "old_instance_type_id": 1,
            "source_compute": "compute1",
            "source_node": "node1",
            "status": "Done",
            "updated_at": "2013-11-50T13:42:02.000000"
        }

        if self.api_version >= api_versions.APIVersion("2.23"):
            migration1.update({"migration_type": "live-migration"})
            migration2.update({"migration_type": "live-migration"})

        if self.api_version >= api_versions.APIVersion("2.59"):
            migration1.update({"uuid": "11111111-07d5-11e1-90e3-e3dffe0c5983"})
            migration2.update({"uuid": "22222222-07d5-11e1-90e3-e3dffe0c5983"})

        if self.api_version >= api_versions.APIVersion("2.80"):
            migration1.update({
                "project_id": "b59c18e5fa284fd384987c5cb25a1853",
                "user_id": "13cc0930d27c4be0acc14d7c47a3e1f7"})
            migration2.update({
                "project_id": "b59c18e5fa284fd384987c5cb25a1853",
                "user_id": "13cc0930d27c4be0acc14d7c47a3e1f7"})

        migration_list = []
        instance_uuid = kw.get('instance_uuid', None)
        if instance_uuid == migration1['instance_uuid']:
            migration_list.append(migration1)
        elif instance_uuid == migration2['instance_uuid']:
            migration_list.append(migration2)
        elif instance_uuid is None:
            migration_list.extend([migration1, migration2])

        migrations = {'migrations': migration_list}

        return (200, FAKE_RESPONSE_HEADERS, migrations)

    #
    # Server Groups
    #

    def get_os_server_groups(self, **kw):
        server_groups = [
            {"members": [], "metadata": {},
             "id": "2cbd51f4-fafe-4cdb-801b-cf913a6f288b",
             "policies": [], "name": "ig1"},
            {"members": [], "metadata": {},
             "id": "4473bb03-4370-4bfb-80d3-dc8cffc47d94",
             "policies": ["anti-affinity"], "name": "ig2"},
            {"members": [], "metadata": {"key": "value"},
             "id": "31ab9bdb-55e1-4ac3-b094-97eeb1b65cc4",
             "policies": [], "name": "ig3"},
            {"members": ["2dccb4a1-02b9-482a-aa23-5799490d6f5d"],
             "metadata": {},
             "id": "4890bb03-7070-45fb-8453-d34556c87d94",
             "policies": ["anti-affinity"], "name": "ig2"}]

        other_project_server_groups = [
            {"members": [], "metadata": {},
             "id": "11111111-1111-1111-1111-111111111111",
             "policies": [], "name": "ig4"},
            {"members": [], "metadata": {},
             "id": "22222222-2222-2222-2222-222222222222",
             "policies": ["anti-affinity"], "name": "ig5"},
            {"members": [], "metadata": {"key": "value"},
             "id": "31ab9bdb-55e1-4ac3-b094-97eeb1b65cc4",
             "policies": [], "name": "ig6"},
            {"members": ["33333333-3333-3333-3333-333333333333"],
             "metadata": {},
             "id": "44444444-4444-4444-4444-444444444444",
             "policies": ["anti-affinity"], "name": "ig5"}]

        if kw.get("all_projects", False):
            server_groups.extend(other_project_server_groups)
        limit = int(kw.get("limit", 1000))
        offset = int(kw.get("offset", 0))
        server_groups = server_groups[offset:limit + 1]
        return (200, {}, {"server_groups": server_groups})

    def _return_server_group(self):
        if self.api_version < api_versions.APIVersion("2.64"):
            r = {'server_group':
                 self.get_os_server_groups()[2]['server_groups'][0]}
        else:
            r = {"members": [], "id": "2cbd51f4-fafe-4cdb-801b-cf913a6f288b",
                 'server_group': {'name': 'ig1', 'policy': 'anti-affinity',
                                  'rules': {'max_server_per_host': 3}}}
        return (200, {}, r)

    def post_os_server_groups(self, body, **kw):
        return self._return_server_group()

    def post_servers_1234_migrations_1_action(self, body):
        return (202, {}, None)

    @api_versions.wraps(start_version="2.23")
    def get_servers_1234_migrations_1(self, **kw):
        migration = {"migration": {
            "created_at": "2016-01-29T13:42:02.000000",
            "dest_compute": "compute2",
            "dest_host": "1.2.3.4",
            "dest_node": "node2",
            "id": 1,
            "server_uuid": "4cfba335-03d8-49b2-8c52-e69043d1e8fe",
            "source_compute": "compute1",
            "source_node": "node1",
            "status": "running",
            "memory_total_bytes": 123456,
            "memory_processed_bytes": 12345,
            "memory_remaining_bytes": 120000,
            "disk_total_bytes": 234567,
            "disk_processed_bytes": 23456,
            "disk_remaining_bytes": 230000,
            "updated_at": "2016-01-29T13:42:02.000000"
        }}

        if self.api_version >= api_versions.APIVersion("2.80"):
            migration['migration'].update({
                "project_id": "b59c18e5fa284fd384987c5cb25a1853",
                "user_id": "13cc0930d27c4be0acc14d7c47a3e1f7"})

        return (200, FAKE_RESPONSE_HEADERS, migration)

    @api_versions.wraps(start_version="2.23")
    def get_servers_1234_migrations(self, **kw):
        migrations = {'migrations': [
            {
                "created_at": "2016-01-29T13:42:02.000000",
                "dest_compute": "compute2",
                "dest_host": "1.2.3.4",
                "dest_node": "node2",
                "id": 1,
                "server_uuid": "4cfba335-03d8-49b2-8c52-e69043d1e8fe",
                "source_compute": "compute1",
                "source_node": "node1",
                "status": "running",
                "memory_total_bytes": 123456,
                "memory_processed_bytes": 12345,
                "memory_remaining_bytes": 120000,
                "disk_total_bytes": 234567,
                "disk_processed_bytes": 23456,
                "disk_remaining_bytes": 230000,
                "updated_at": "2016-01-29T13:42:02.000000"
            }]}

        if self.api_version >= api_versions.APIVersion("2.80"):
            migrations['migrations'][0].update({
                "project_id": "b59c18e5fa284fd384987c5cb25a1853",
                "user_id": "13cc0930d27c4be0acc14d7c47a3e1f7"})

        return (200, FAKE_RESPONSE_HEADERS, migrations)

    def delete_servers_1234_migrations_1(self):
        return (202, {}, None)

    def put_servers_1234(self, **kw):
        return (201, {}, None)

    def put_servers_1234_tags_tag(self, **kw):
        return (201, {}, None)

    def put_servers_1234_tags_tag1(self, **kw):
        return (201, {}, None)

    def put_servers_1234_tags_tag2(self, **kw):
        return (201, {}, None)

    def put_servers_1234_tags_tag3(self, **kw):
        return (201, {}, None)

    def put_servers_1234_tags(self, **kw):
        return (201, {}, None)

    def get_servers_1234_tags(self, **kw):
        return (200, {}, {'tags': ['tag1', 'tag2']})

    def delete_servers_1234_tags_tag(self, **kw):
        return (204, {}, None)

    def delete_servers_1234_tags_tag1(self, **kw):
        return (204, {}, None)

    def delete_servers_1234_tags_tag2(self, **kw):
        return (204, {}, None)

    def delete_servers_1234_tags_tag3(self, **kw):
        return (204, {}, None)

    def delete_servers_1234_tags(self, **kw):
        return (204, {}, None)

    def post_os_assisted_volume_snapshots(self, **kw):
        return (202, FAKE_RESPONSE_HEADERS,
                {'snapshot': {'id': 'blah', 'volumeId': '1'}})

    def delete_os_assisted_volume_snapshots_x(self, **kw):
        return (202, FAKE_RESPONSE_HEADERS, {})

    def post_os_server_external_events(self, **kw):
        return (200, FAKE_RESPONSE_HEADERS, {
            'events': [
                {'name': 'test-event',
                 'status': 'completed',
                 'tag': 'tag',
                 'server_uuid': 'fake-uuid1'},
                {'name': 'test-event',
                 'status': 'completed',
                 'tag': 'tag',
                 'server_uuid': 'fake-uuid2'}]})



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient import exceptions
from novaclient.tests.unit.fixture_data import client
from novaclient.tests.unit.fixture_data import hypervisors as data
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes


class HypervisorsTest(utils.FixturedTestCase):

    client_fixture_class = client.V1
    data_fixture_class = data.V1

    def compare_to_expected(self, expected, hyper):
        for key, value in expected.items():
            self.assertEqual(getattr(hyper, key), value)

    def test_hypervisor_index(self):
        expected = [
            dict(id=self.data_fixture.hyper_id_1,
                 hypervisor_hostname='hyper1'),
            dict(id=self.data_fixture.hyper_id_2,
                 hypervisor_hostname='hyper2')]

        result = self.cs.hypervisors.list(False)
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-hypervisors')

        for idx, hyper in enumerate(result):
            self.compare_to_expected(expected[idx], hyper)

    def test_hypervisor_detail(self):
        expected = [
            dict(id=self.data_fixture.hyper_id_1,
                 service=dict(id=self.data_fixture.service_id_1,
                              host='compute1'),
                 vcpus=4,
                 memory_mb=10 * 1024,
                 local_gb=250,
                 vcpus_used=2,
                 memory_mb_used=5 * 1024,
                 local_gb_used=125,
                 hypervisor_type="xen",
                 hypervisor_version=3,
                 hypervisor_hostname="hyper1",
                 free_ram_mb=5 * 1024,
                 free_disk_gb=125,
                 current_workload=2,
                 running_vms=2,
                 cpu_info='cpu_info',
                 disk_available_least=100,
                 state='up',
                 status='enabled'),
            dict(id=self.data_fixture.hyper_id_2,
                 service=dict(id=self.data_fixture.service_id_2,
                              host="compute2"),
                 vcpus=4,
                 memory_mb=10 * 1024,
                 local_gb=250,
                 vcpus_used=2,
                 memory_mb_used=5 * 1024,
                 local_gb_used=125,
                 hypervisor_type="xen",
                 hypervisor_version=3,
                 hypervisor_hostname="hyper2",
                 free_ram_mb=5 * 1024,
                 free_disk_gb=125,
                 current_workload=2,
                 running_vms=2,
                 cpu_info='cpu_info',
                 disk_available_least=100,
                 state='up',
                 status='enabled')]

        if self.cs.api_version >= api_versions.APIVersion('2.88'):
            for hypervisor in expected:
                del hypervisor['current_workload']
                del hypervisor['disk_available_least']
                del hypervisor['free_ram_mb']
                del hypervisor['free_disk_gb']
                del hypervisor['local_gb']
                del hypervisor['local_gb_used']
                del hypervisor['memory_mb']
                del hypervisor['memory_mb_used']
                del hypervisor['running_vms']
                del hypervisor['vcpus']
                del hypervisor['vcpus_used']
                hypervisor['uptime'] = 'fake uptime'

        result = self.cs.hypervisors.list()
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-hypervisors/detail')

        for idx, hyper in enumerate(result):
            self.compare_to_expected(expected[idx], hyper)

    def test_hypervisor_search(self):
        expected = [
            dict(id=self.data_fixture.hyper_id_1,
                 hypervisor_hostname='hyper1',
                 state='up',
                 status='enabled'),
            dict(id=self.data_fixture.hyper_id_2,
                 hypervisor_hostname='hyper2',
                 state='up',
                 status='enabled')]

        result = self.cs.hypervisors.search('hyper')
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        if self.cs.api_version >= api_versions.APIVersion('2.53'):
            self.assert_called(
                'GET', '/os-hypervisors?hypervisor_hostname_pattern=hyper')
        else:
            self.assert_called('GET', '/os-hypervisors/hyper/search')

        for idx, hyper in enumerate(result):
            self.compare_to_expected(expected[idx], hyper)

    def test_hypervisor_search_unicode(self):
        hypervisor_match = '\\u5de5\\u4f5c'
        if self.cs.api_version >= api_versions.APIVersion('2.53'):
            self.assertRaises(exceptions.BadRequest,
                              self.cs.hypervisors.search,
                              hypervisor_match)
        else:
            self.assertRaises(exceptions.NotFound,
                              self.cs.hypervisors.search,
                              hypervisor_match)

    def test_hypervisor_search_detailed(self):
        # detailed=True is not supported before 2.53
        ex = self.assertRaises(exceptions.UnsupportedVersion,
                               self.cs.hypervisors.search, 'hyper',
                               detailed=True)
        self.assertIn('Parameter "detailed" requires API version 2.53 or '
                      'greater.', str(ex))

    def test_hypervisor_servers(self):
        expected = [
            dict(id=self.data_fixture.hyper_id_1,
                 hypervisor_hostname='hyper1',
                 state='up',
                 status='enabled',
                 servers=[
                     dict(name='inst1', uuid='uuid1'),
                     dict(name='inst2', uuid='uuid2')]),
            dict(id=self.data_fixture.hyper_id_2,
                 hypervisor_hostname='hyper2',
                 state='up',
                 status='enabled',
                 servers=[
                     dict(name='inst3', uuid='uuid3'),
                     dict(name='inst4', uuid='uuid4')]),
        ]

        result = self.cs.hypervisors.search('hyper', True)
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        if self.cs.api_version >= api_versions.APIVersion('2.53'):
            self.assert_called(
                'GET', '/os-hypervisors?hypervisor_hostname_pattern=hyper&'
                       'with_servers=True')
        else:
            self.assert_called('GET', '/os-hypervisors/hyper/servers')

        for idx, hyper in enumerate(result):
            self.compare_to_expected(expected[idx], hyper)

    def test_hypervisor_get(self):
        expected = dict(
            id=self.data_fixture.hyper_id_1,
            service=dict(id=self.data_fixture.service_id_1, host='compute1'),
            vcpus=4,
            memory_mb=10 * 1024,
            local_gb=250,
            vcpus_used=2,
            memory_mb_used=5 * 1024,
            local_gb_used=125,
            hypervisor_type="xen",
            hypervisor_version=3,
            hypervisor_hostname="hyper1",
            free_ram_mb=5 * 1024,
            free_disk_gb=125,
            current_workload=2,
            running_vms=2,
            cpu_info='cpu_info',
            disk_available_least=100,
            state='up',
            status='enabled')

        if self.cs.api_version >= api_versions.APIVersion('2.88'):
            del expected['current_workload']
            del expected['disk_available_least']
            del expected['free_ram_mb']
            del expected['free_disk_gb']
            del expected['local_gb']
            del expected['local_gb_used']
            del expected['memory_mb']
            del expected['memory_mb_used']
            del expected['running_vms']
            del expected['vcpus']
            del expected['vcpus_used']
            expected['uptime'] = 'fake uptime'

        result = self.cs.hypervisors.get(self.data_fixture.hyper_id_1)
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called(
            'GET', '/os-hypervisors/%s' % self.data_fixture.hyper_id_1)

        self.compare_to_expected(expected, result)

    def test_hypervisor_uptime(self):
        expected = dict(
            id=self.data_fixture.hyper_id_1,
            hypervisor_hostname="hyper1",
            uptime="fake uptime",
            state='up',
            status='enabled')

        result = self.cs.hypervisors.uptime(self.data_fixture.hyper_id_1)
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called(
            'GET', '/os-hypervisors/%s/uptime' % self.data_fixture.hyper_id_1)

        self.compare_to_expected(expected, result)

    def test_hypervisor_statistics(self):
        expected = dict(
            count=2,
            vcpus=8,
            memory_mb=20 * 1024,
            local_gb=500,
            vcpus_used=4,
            memory_mb_used=10 * 1024,
            local_gb_used=250,
            free_ram_mb=10 * 1024,
            free_disk_gb=250,
            current_workload=4,
            running_vms=4,
            disk_available_least=200,
        )

        result = self.cs.hypervisors.statistics()
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-hypervisors/statistics')

        self.compare_to_expected(expected, result)

        # Test for Bug #1370415, the line below used to raise AttributeError
        self.assertEqual("<HypervisorStats: 2 Hypervisors>",
                         result.__repr__())


class HypervisorsV233Test(HypervisorsTest):
    def setUp(self):
        super(HypervisorsV233Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.33")

    def test_use_limit_marker_params(self):
        params = {'limit': '10', 'marker': 'fake-marker'}
        self.cs.hypervisors.list(**params)
        for k, v in params.items():
            self.assertEqual([v], self.requests_mock.last_request.qs[k])


class HypervisorsV253Test(HypervisorsV233Test):
    """Tests the os-hypervisors 2.53 API bindings."""
    data_fixture_class = data.V253

    def setUp(self):
        super(HypervisorsV253Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion("2.53")

    def test_hypervisor_search_detailed(self):
        expected = [
            dict(id=self.data_fixture.hyper_id_1,
                 state='up',
                 status='enabled',
                 hypervisor_hostname='hyper1'),
            dict(id=self.data_fixture.hyper_id_2,
                 state='up',
                 status='enabled',
                 hypervisor_hostname='hyper2')]
        result = self.cs.hypervisors.search('hyper', detailed=True)
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called(
            'GET', '/os-hypervisors/detail?hypervisor_hostname_pattern=hyper')
        for idx, hyper in enumerate(result):
            self.compare_to_expected(expected[idx], hyper)


class HypervisorsV288Test(HypervisorsV253Test):
    data_fixture_class = data.V288

    def setUp(self):
        super().setUp()
        self.cs.api_version = api_versions.APIVersion('2.88')

    def test_hypervisor_uptime(self):
        expected = {
            'id': self.data_fixture.hyper_id_1,
            'hypervisor_hostname': 'hyper1',
            'uptime': 'fake uptime',
            'state': 'up',
            'status': 'enabled',
        }

        result = self.cs.hypervisors.uptime(self.data_fixture.hyper_id_1)
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called(
            'GET', '/os-hypervisors/%s' % self.data_fixture.hyper_id_1)

        self.compare_to_expected(expected, result)

    def test_hypervisor_statistics(self):
        exc = self.assertRaises(
            exceptions.UnsupportedVersion,
            self.cs.hypervisor_stats.statistics)
        self.assertIn(
            "The 'statistics' API is removed in API version 2.88 or later.",
            str(exc))



#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import pbr.version

from novaclient import api_versions


__version__ = pbr.version.VersionInfo('python-novaclient').version_string()

API_MIN_VERSION = api_versions.APIVersion("2.1")
# The max version should be the latest version that is supported in the client,
# not necessarily the latest that the server can provide. This is only bumped
# when client supported the max version, and bumped sequentially, otherwise
# the client may break due to server side new version may include some
# backward incompatible change.
API_MAX_VERSION = api_versions.APIVersion("2.96")



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Network interface.
"""
from novaclient import base
from novaclient import exceptions
from novaclient.i18n import _


class Network(base.Resource):
    """
    A network as defined in the Networking (Neutron) API.
    """
    HUMAN_ID = True
    NAME_ATTR = "name"

    def __repr__(self):
        return "<Network: %s>" % self.name


class NeutronManager(base.Manager):
    """A manager for name -> id lookups for neutron networks.

    This uses neutron directly from service catalog. Do not use it
    for anything else besides that. You have been warned.
    """

    resource_class = Network

    def find_network(self, name):
        """Find a network by name (user provided input)."""

        with self.alternate_service_type(
                'network', allowed_types=('network',)):

            matches = self._list('/v2.0/networks?name=%s' % name, 'networks')

            num_matches = len(matches)
            if num_matches == 0:
                msg = "No %s matching %s." % (
                    self.resource_class.__name__, name)
                raise exceptions.NotFound(404, msg)
            elif num_matches > 1:
                msg = (_("Multiple %(class)s matches found for "
                         "'%(name)s', use an ID to be more specific.") %
                       {'class': self.resource_class.__name__.lower(),
                        'name': name})
                raise exceptions.NoUniqueMatch(msg)
            else:
                matches[0].append_request_ids(matches.request_ids)
                return matches[0]



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Keypair interface
"""

from novaclient import api_versions
from novaclient import base


class Keypair(base.Resource):
    """
    A keypair is a ssh key that can be injected into a server on launch.
    """

    def __repr__(self):
        return "<Keypair: %s>" % self.id

    def _add_details(self, info):
        dico = 'keypair' in info and \
            info['keypair'] or info
        for (k, v) in dico.items():
            # NOTE(rpodolyaka): keypair name allows us to uniquely identify
            #                   a specific keypair, while its id attribute
            #                   is nothing more than an implementation
            #                   detail. We can safely omit the id attribute
            #                   here to ensure setattr() won't raise
            #                   AttributeError trying to set read-only
            #                   property id
            if k != 'id':
                setattr(self, k, v)

    @property
    def id(self):
        return self.name

    def delete(self):
        """
        Delete this keypair.

        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self.manager.delete(self)


class KeypairManager(base.ManagerWithFind):
    resource_class = Keypair
    keypair_prefix = "os-keypairs"
    is_alphanum_id_allowed = True

    @api_versions.wraps("2.0", "2.9")
    def get(self, keypair):
        """
        Get a keypair.

        :param keypair: The ID of the keypair to get.
        :rtype: :class:`Keypair`
        """
        return self._get("/%s/%s" % (self.keypair_prefix, base.getid(keypair)),
                         "keypair")

    @api_versions.wraps("2.10")
    def get(self, keypair, user_id=None):
        """
        Get a keypair.

        :param keypair: The ID of the keypair to get.
        :param user_id: Id of key-pair owner (Admin only).
        :rtype: :class:`Keypair`
        """
        query_string = "?user_id=%s" % user_id if user_id else ""
        url = "/%s/%s%s" % (self.keypair_prefix, base.getid(keypair),
                            query_string)
        return self._get(url, "keypair")

    @api_versions.wraps("2.0", "2.1")
    def create(self, name, public_key=None):
        """
        Create a keypair

        :param name: name for the keypair to create
        :param public_key: existing public key to import
        """
        body = {'keypair': {'name': name}}
        if public_key:
            body['keypair']['public_key'] = public_key
        return self._create('/%s' % self.keypair_prefix, body, 'keypair')

    @api_versions.wraps("2.2", "2.9")
    def create(self, name, public_key=None, key_type="ssh"):
        """
        Create a keypair

        :param name: name for the keypair to create
        :param public_key: existing public key to import
        :param key_type: keypair type to create
        """
        body = {'keypair': {'name': name,
                            'type': key_type}}
        if public_key:
            body['keypair']['public_key'] = public_key
        return self._create('/%s' % self.keypair_prefix, body, 'keypair')

    @api_versions.wraps("2.10", "2.91")
    def create(self, name, public_key=None, key_type="ssh", user_id=None):
        """
        Create a keypair

        :param name: name for the keypair to create
        :param public_key: existing public key to import
        :param key_type: keypair type to create
        :param user_id: user to add.
        """
        body = {'keypair': {'name': name,
                            'type': key_type}}
        if public_key:
            body['keypair']['public_key'] = public_key
        if user_id:
            body['keypair']['user_id'] = user_id
        return self._create('/%s' % self.keypair_prefix, body, 'keypair')

    @api_versions.wraps("2.92")
    def create(self, name, public_key, key_type="ssh", user_id=None):
        """
        Create a keypair

        :param name: name for the keypair to create
        :param public_key: existing public key to import
        :param key_type: keypair type to create
        :param user_id: user to add.
        """
        body = {'keypair': {'name': name,
                            'type': key_type,
                            'public_key': public_key}}
        if user_id:
            body['keypair']['user_id'] = user_id
        return self._create('/%s' % self.keypair_prefix, body, 'keypair')

    @api_versions.wraps("2.0", "2.9")
    def delete(self, key):
        """
        Delete a keypair

        :param key: The :class:`Keypair` (or its ID) to delete.
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        return self._delete('/%s/%s' % (self.keypair_prefix, base.getid(key)))

    @api_versions.wraps("2.10")
    def delete(self, key, user_id=None):
        """
        Delete a keypair

        :param key: The :class:`Keypair` (or its ID) to delete.
        :param user_id: Id of key-pair owner (Admin only).
        :returns: An instance of novaclient.base.TupleWithMeta
        """
        query_string = "?user_id=%s" % user_id if user_id else ""
        url = '/%s/%s%s' % (self.keypair_prefix, base.getid(key), query_string)
        return self._delete(url)

    @api_versions.wraps("2.0", "2.9")
    def list(self):
        """
        Get a list of keypairs.
        """
        return self._list('/%s' % self.keypair_prefix, 'keypairs')

    @api_versions.wraps("2.10", "2.34")
    def list(self, user_id=None):
        """
        Get a list of keypairs.

        :param user_id: Id of key-pairs owner (Admin only).
        """
        params = {}
        if user_id:
            params['user_id'] = user_id
        return self._list('/%s' % self.keypair_prefix, 'keypairs',
                          filters=params)

    @api_versions.wraps("2.35")
    def list(self, user_id=None, marker=None, limit=None):
        """
        Get a list of keypairs.

        :param user_id: Id of key-pairs owner (Admin only).
        :param marker: Begin returning keypairs that appear later in the
                       keypair list than that represented by this keypair name
                       (optional).
        :param limit: maximum number of keypairs to return (optional).
                      Note the API server has a configurable default limit.
                      If no limit is specified here or limit is larger than
                      default, the default limit will be used.
        """
        params = {}
        if user_id:
            params['user_id'] = user_id
        if limit:
            params['limit'] = int(limit)
        if marker:
            params['marker'] = str(marker)
        return self._list('/%s' % self.keypair_prefix, 'keypairs',
                          filters=params)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import base
from novaclient import utils


class Extension(base.HookableMixin):
    """Extension descriptor."""

    SUPPORTED_HOOKS = ('__pre_parse_args__', '__post_parse_args__')

    def __init__(self, name, module):
        self.name = name
        self.module = module
        self._parse_extension_module()

    def _parse_extension_module(self):
        self.manager_class = None
        for attr_name, attr_value in self.module.__dict__.items():
            if attr_name in self.SUPPORTED_HOOKS:
                self.add_hook(attr_name, attr_value)
            elif utils.safe_issubclass(attr_value, base.Manager):
                self.manager_class = attr_value

    def __repr__(self):
        return "<Extension '%s'>" % self.name



# Copyright 2012 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Service interface.
"""

from urllib import parse

from novaclient import api_versions
from novaclient import base


class Service(base.Resource):
    def __repr__(self):
        return "<Service: %s>" % self.id

    def _add_details(self, info):
        dico = 'resource' in info and info['resource'] or info
        for (k, v) in dico.items():
            setattr(self, k, v)


class ServiceManager(base.ManagerWithFind):
    resource_class = Service

    def list(self, host=None, binary=None):
        """
        Get a list of services.

        :param host: destination host name.
        """
        url = "/os-services"
        filters = []
        if host:
            filters.append(("host", host))
        if binary:
            filters.append(("binary", binary))
        if filters:
            url = "%s?%s" % (url, parse.urlencode(filters))
        return self._list(url, "services")

    @api_versions.wraps("2.0", "2.10")
    def _update_body(self, host, binary, disabled_reason=None):
        body = {"host": host,
                "binary": binary}
        if disabled_reason is not None:
            body["disabled_reason"] = disabled_reason
        return body

    @api_versions.wraps("2.11")
    def _update_body(self, host, binary, disabled_reason=None,
                     force_down=None):
        body = {"host": host,
                "binary": binary}
        if disabled_reason is not None:
            body["disabled_reason"] = disabled_reason
        if force_down is not None:
            body["forced_down"] = force_down
        return body

    @api_versions.wraps('2.0', '2.52')
    def enable(self, host, binary):
        """Enable the service specified by hostname and binary."""
        body = self._update_body(host, binary)
        return self._update("/os-services/enable", body, "service")

    @api_versions.wraps('2.53')
    def enable(self, service_uuid):
        """Enable the service specified by the service UUID ID.

        :param service_uuid: The UUID ID of the service to enable.
        """
        return self._update(
            "/os-services/%s" % service_uuid, {'status': 'enabled'}, "service")

    @api_versions.wraps('2.0', '2.52')
    def disable(self, host, binary):
        """Disable the service specified by hostname and binary."""
        body = self._update_body(host, binary)
        return self._update("/os-services/disable", body, "service")

    @api_versions.wraps('2.53')
    def disable(self, service_uuid):
        """Disable the service specified by the service UUID ID.

        :param service_uuid: The UUID ID of the service to disable.
        """
        return self._update("/os-services/%s" % service_uuid,
                            {'status': 'disabled'}, "service")

    @api_versions.wraps('2.0', '2.52')
    def disable_log_reason(self, host, binary, reason):
        """Disable the service with reason."""
        body = self._update_body(host, binary, reason)
        return self._update("/os-services/disable-log-reason", body, "service")

    @api_versions.wraps('2.53')
    def disable_log_reason(self, service_uuid, reason):
        """Disable the service with a reason.

        :param service_uuid: The UUID ID of the service to disable.
        :param reason: The reason for disabling a service. The minimum length
            is 1 and the maximum length is 255.
        """
        body = {
            'status': 'disabled',
            'disabled_reason': reason
        }
        return self._update("/os-services/%s" % service_uuid, body, "service")

    def delete(self, service_id):
        """Delete a service.

        :param service_id: Before microversion 2.53, this must be an integer id
            and may not uniquely the service in a multi-cell deployment.
            Starting with microversion 2.53 this must be a UUID.
        """
        return self._delete("/os-services/%s" % service_id)

    @api_versions.wraps("2.11", "2.52")
    def force_down(self, host, binary, force_down=None):
        """Force service state to down specified by hostname and binary."""
        body = self._update_body(host, binary, force_down=force_down)
        return self._update("/os-services/force-down", body, "service")

    @api_versions.wraps("2.53")
    def force_down(self, service_uuid, force_down):
        """Update the service's ``forced_down`` field specified by the
        service UUID ID.

        :param service_uuid: The UUID ID of the service.
        :param force_down: Whether or not this service was forced down manually
            by an administrator. This value is useful to know that some 3rd
            party has verified the service should be marked down.
        """
        return self._update("/os-services/%s" % service_uuid,
                            {'forced_down': force_down}, "service")



# Copyright 2012 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient.tests.unit.fixture_data import agents as data
from novaclient.tests.unit.fixture_data import client
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes
from novaclient.v2 import agents


class AgentsTest(utils.FixturedTestCase):

    data_fixture_class = data.Fixture

    scenarios = [('original', {'client_fixture_class': client.V1}),
                 ('session', {'client_fixture_class': client.SessionV1})]

    def stub_hypervisors(self, hypervisor='kvm'):
        get_os_agents = {
            'agents': [
                {
                    'hypervisor': hypervisor,
                    'os': 'win',
                    'architecture': 'x86',
                    'version': '7.0',
                    'url': 'xxx://xxxx/xxx/xxx',
                    'md5hash': 'add6bb58e139be103324d04d82d8f545',
                    'id': 1
                },
                {
                    'hypervisor': hypervisor,
                    'os': 'linux',
                    'architecture': 'x86',
                    'version': '16.0',
                    'url': 'xxx://xxxx/xxx/xxx1',
                    'md5hash': 'add6bb58e139be103324d04d82d8f546',
                    'id': 2
                },
            ]
        }

        headers = {'Content-Type': 'application/json',
                   'x-openstack-request-id': fakes.FAKE_REQUEST_ID}
        self.requests_mock.get(self.data_fixture.url(),
                               json=get_os_agents,
                               headers=headers)

    def test_list_agents(self):
        self.stub_hypervisors()
        ags = self.cs.agents.list()
        self.assert_called('GET', '/os-agents')
        self.assert_request_id(ags, fakes.FAKE_REQUEST_ID_LIST)
        for a in ags:
            self.assertIsInstance(a, agents.Agent)
            self.assertEqual('kvm', a.hypervisor)

    def test_list_agents_with_hypervisor(self):
        self.stub_hypervisors('xen')
        ags = self.cs.agents.list('xen')
        self.assert_called('GET', '/os-agents?hypervisor=xen')
        self.assert_request_id(ags, fakes.FAKE_REQUEST_ID_LIST)
        for a in ags:
            self.assertIsInstance(a, agents.Agent)
            self.assertEqual('xen', a.hypervisor)

    def test_agents_create(self):
        ag = self.cs.agents.create('win', 'x86', '7.0',
                                   '/xxx/xxx/xxx',
                                   'add6bb58e139be103324d04d82d8f546',
                                   'xen')
        self.assert_request_id(ag, fakes.FAKE_REQUEST_ID_LIST)
        body = {'agent': {'url': '/xxx/xxx/xxx',
                          'hypervisor': 'xen',
                          'md5hash': 'add6bb58e139be103324d04d82d8f546',
                          'version': '7.0',
                          'architecture': 'x86',
                          'os': 'win'}}
        self.assert_called('POST', '/os-agents', body)
        self.assertEqual(1, ag._info.copy()['id'])

    def test_agents_delete(self):
        ret = self.cs.agents.delete('1')
        self.assert_request_id(ret, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('DELETE', '/os-agents/1')

    def _build_example_update_body(self):
        return {"para": {
            "url": "/yyy/yyyy/yyyy",
            "version": "8.0",
            "md5hash": "add6bb58e139be103324d04d82d8f546"}}

    def test_agents_modify(self):
        ag = self.cs.agents.update('1', '8.0',
                                   '/yyy/yyyy/yyyy',
                                   'add6bb58e139be103324d04d82d8f546')
        self.assert_request_id(ag, fakes.FAKE_REQUEST_ID_LIST)
        body = self._build_example_update_body()
        self.assert_called('PUT', '/os-agents/1', body)
        self.assertEqual(1, ag.id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils

from novaclient.tests.functional.v2.legacy import test_extended_attributes


class TestExtAttrNovaClientV23(test_extended_attributes.TestExtAttrNovaClient):
    """Functional tests for os-extended-server-attributes, microversion 2.3"""

    COMPUTE_API_VERSION = "2.3"

    def test_extended_server_attributes(self):
        server, volume = self._create_server_and_attach_volume()
        table = self.nova('show %s' % server.id)
        # Check that attributes listed below exist in 'nova show' table and
        # they are exactly Property attributes (not an instance's name, e.g.)
        # The _get_value_from_the_table() will raise an exception
        # if attr is not a key of the table dict (first column)
        for attr in ['OS-EXT-SRV-ATTR:reservation_id',
                     'OS-EXT-SRV-ATTR:launch_index',
                     'OS-EXT-SRV-ATTR:ramdisk_id',
                     'OS-EXT-SRV-ATTR:kernel_id',
                     'OS-EXT-SRV-ATTR:hostname',
                     'OS-EXT-SRV-ATTR:root_device_name']:
            self._get_value_from_the_table(table, attr)
        # Check that attribute given below also exists in 'nova show' table
        # as a key (first column) of table dict
        volume_attr = self._get_value_from_the_table(
            table, 'os-extended-volumes:volumes_attached')
        # Check that 'delete_on_termination' exists as a key
        # of volume_attr dict
        self.assertIn('delete_on_termination', jsonutils.loads(volume_attr)[0])



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import api_versions
from novaclient import exceptions
from novaclient.tests.unit.fixture_data import aggregates as data
from novaclient.tests.unit.fixture_data import client
from novaclient.tests.unit import utils
from novaclient.tests.unit.v2 import fakes
from novaclient.v2 import aggregates
from novaclient.v2 import images


class AggregatesTest(utils.FixturedTestCase):

    data_fixture_class = data.Fixture

    scenarios = [('original', {'client_fixture_class': client.V1}),
                 ('session', {'client_fixture_class': client.SessionV1})]

    def test_list_aggregates(self):
        result = self.cs.aggregates.list()
        self.assert_request_id(result, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-aggregates')
        for aggregate in result:
            self.assertIsInstance(aggregate, aggregates.Aggregate)

    def test_create_aggregate(self):
        body = {"aggregate": {"name": "test", "availability_zone": "nova1"}}
        aggregate = self.cs.aggregates.create("test", "nova1")
        self.assert_request_id(aggregate, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates', body)
        self.assertIsInstance(aggregate, aggregates.Aggregate)

    def test_get(self):
        aggregate = self.cs.aggregates.get("1")
        self.assert_request_id(aggregate, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-aggregates/1')
        self.assertIsInstance(aggregate, aggregates.Aggregate)

        aggregate2 = self.cs.aggregates.get(aggregate)
        self.assert_request_id(aggregate2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-aggregates/1')
        self.assertIsInstance(aggregate2, aggregates.Aggregate)

    def test_get_details(self):
        aggregate = self.cs.aggregates.get_details("1")
        self.assert_request_id(aggregate, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-aggregates/1')
        self.assertIsInstance(aggregate, aggregates.Aggregate)

        aggregate2 = self.cs.aggregates.get_details(aggregate)
        self.assert_request_id(aggregate2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('GET', '/os-aggregates/1')
        self.assertIsInstance(aggregate2, aggregates.Aggregate)

    def test_update(self):
        aggregate = self.cs.aggregates.get("1")
        values = {"name": "foo"}
        body = {"aggregate": values}

        result1 = aggregate.update(values)
        self.assert_request_id(result1, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('PUT', '/os-aggregates/1', body)
        self.assertIsInstance(result1, aggregates.Aggregate)

        result2 = self.cs.aggregates.update(2, values)
        self.assert_request_id(result2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('PUT', '/os-aggregates/2', body)
        self.assertIsInstance(result2, aggregates.Aggregate)

    def test_update_with_availability_zone(self):
        aggregate = self.cs.aggregates.get("1")
        values = {"name": "foo", "availability_zone": "new_zone"}
        body = {"aggregate": values}

        result3 = self.cs.aggregates.update(aggregate, values)
        self.assert_request_id(result3, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('PUT', '/os-aggregates/1', body)
        self.assertIsInstance(result3, aggregates.Aggregate)

    def test_add_host(self):
        aggregate = self.cs.aggregates.get("1")
        host = "host1"
        body = {"add_host": {"host": "host1"}}

        result1 = aggregate.add_host(host)
        self.assert_request_id(result1, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/1/action', body)
        self.assertIsInstance(result1, aggregates.Aggregate)

        result2 = self.cs.aggregates.add_host("2", host)
        self.assert_request_id(result2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/2/action', body)
        self.assertIsInstance(result2, aggregates.Aggregate)

        result3 = self.cs.aggregates.add_host(aggregate, host)
        self.assert_request_id(result3, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/1/action', body)
        self.assertIsInstance(result3, aggregates.Aggregate)

    def test_remove_host(self):
        aggregate = self.cs.aggregates.get("1")
        host = "host1"
        body = {"remove_host": {"host": "host1"}}

        result1 = aggregate.remove_host(host)
        self.assert_request_id(result1, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/1/action', body)
        self.assertIsInstance(result1, aggregates.Aggregate)

        result2 = self.cs.aggregates.remove_host("2", host)
        self.assert_request_id(result2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/2/action', body)
        self.assertIsInstance(result2, aggregates.Aggregate)

        result3 = self.cs.aggregates.remove_host(aggregate, host)
        self.assert_request_id(result3, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/1/action', body)
        self.assertIsInstance(result3, aggregates.Aggregate)

    def test_set_metadata(self):
        aggregate = self.cs.aggregates.get("1")
        metadata = {"foo": "bar"}
        body = {"set_metadata": {"metadata": metadata}}

        result1 = aggregate.set_metadata(metadata)
        self.assert_request_id(result1, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/1/action', body)
        self.assertIsInstance(result1, aggregates.Aggregate)

        result2 = self.cs.aggregates.set_metadata(2, metadata)
        self.assert_request_id(result2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/2/action', body)
        self.assertIsInstance(result2, aggregates.Aggregate)

        result3 = self.cs.aggregates.set_metadata(aggregate, metadata)
        self.assert_request_id(result3, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('POST', '/os-aggregates/1/action', body)
        self.assertIsInstance(result3, aggregates.Aggregate)

    def test_delete_aggregate(self):
        aggregate = self.cs.aggregates.list()[0]
        result1 = aggregate.delete()
        self.assert_request_id(result1, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('DELETE', '/os-aggregates/1')

        result2 = self.cs.aggregates.delete('1')
        self.assert_request_id(result2, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('DELETE', '/os-aggregates/1')

        result3 = self.cs.aggregates.delete(aggregate)
        self.assert_request_id(result3, fakes.FAKE_REQUEST_ID_LIST)
        self.assert_called('DELETE', '/os-aggregates/1')


class AggregatesV281Test(utils.FixturedTestCase):
    api_version = "2.81"
    data_fixture_class = data.Fixture

    scenarios = [('original', {'client_fixture_class': client.V1}),
                 ('session', {'client_fixture_class': client.SessionV1})]

    def setUp(self):
        super(AggregatesV281Test, self).setUp()
        self.cs.api_version = api_versions.APIVersion(self.api_version)

    def test_cache_images(self):
        aggregate = self.cs.aggregates.list()[0]
        _images = [images.Image(self.cs.aggregates, {'id': '1'}),
                   images.Image(self.cs.aggregates, {'id': '2'})]
        aggregate.cache_images(_images)
        expected_body = {'cache': [{'id': image.id}
                                   for image in _images]}
        self.assert_called('POST', '/os-aggregates/1/images',
                           expected_body)

    def test_cache_images_just_ids(self):
        aggregate = self.cs.aggregates.list()[0]
        _images = ['1']
        aggregate.cache_images(_images)
        expected_body = {'cache': [{'id': '1'}]}
        self.assert_called('POST', '/os-aggregates/1/images',
                           expected_body)

    def test_cache_images_pre281(self):
        self.cs.api_version = api_versions.APIVersion('2.80')
        aggregate = self.cs.aggregates.list()[0]
        _images = [images.Image(self.cs.aggregates, {'id': '1'})]
        self.assertRaises(exceptions.VersionNotFoundForAPIMethod,
                          aggregate.cache_images, _images)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_utils import timeutils

from novaclient.tests.functional import base


class TestInstanceUsageAuditLogCLI(base.ClientTestBase):
    COMPUTE_API_VERSION = '2.1'

    # NOTE(takashin): By default, 'instance_usage_audit' is False in nova.
    # So the instance usage audit log is not recorded.
    # Therefore an empty result can be got.
    # But it is tested here to call APIs and get responses normally.

    @staticmethod
    def _get_begin_end_time():
        current = timeutils.utcnow()

        end = datetime.datetime(day=1, month=current.month, year=current.year)
        year = end.year

        if current.month == 1:
            year -= 1
            month = 12
        else:
            month = current.month - 1

        begin = datetime.datetime(day=1, month=month, year=year)

        return (begin, end)

    def test_get_os_instance_usage_audit_log(self):
        (begin, end) = self._get_begin_end_time()
        expected = {
            'hosts_not_run': '[]',
            'log': '{}',
            'num_hosts': '0',
            'num_hosts_done': '0',
            'num_hosts_not_run': '0',
            'num_hosts_running': '0',
            'overall_status': 'ALL hosts done. 0 errors.',
            'total_errors': '0',
            'total_instances': '0',
            'period_beginning': str(begin),
            'period_ending': str(end)
        }

        output = self.nova('instance-usage-audit-log')

        for key in expected.keys():
            self.assertEqual(expected[key],
                             self._get_value_from_the_table(output, key))

    def test_get_os_instance_usage_audit_log_with_before(self):
        expected = {
            'hosts_not_run': '[]',
            'log': '{}',
            'num_hosts': '0',
            'num_hosts_done': '0',
            'num_hosts_not_run': '0',
            'num_hosts_running': '0',
            'overall_status': 'ALL hosts done. 0 errors.',
            'total_errors': '0',
            'total_instances': '0',
            'period_beginning': '2016-11-01 00:00:00',
            'period_ending': '2016-12-01 00:00:00'
        }

        output = self.nova(
            'instance-usage-audit-log --before "2016-12-10 13:59:59.999999"')

        for key in expected.keys():
            self.assertEqual(expected[key],
                             self._get_value_from_the_table(output, key))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from novaclient.tests.functional import base


class TestServersAPI(base.ClientTestBase):
    def test_server_ips(self):
        server_name = "test_server"
        initial_server = self.client.servers.create(
            server_name, self.image, self.flavor,
            nics=[{"net-id": self.network.id}])
        self.addCleanup(initial_server.delete)

        for x in range(60):
            server = self.client.servers.get(initial_server)
            if server.status == "ACTIVE":
                break
            else:
                time.sleep(1)
        else:
            self.fail("Server %s did not go ACTIVE after 60s" % server)

        ips = self.client.servers.ips(server)
        self.assertIn(self.network.name, ips)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient.tests.functional.v2.legacy import test_hypervisors


class TestHypervisorsV28(test_hypervisors.TestHypervisors):

    COMPUTE_API_VERSION = "2.28"

    def test_list(self):
        self._test_list(dict)


class TestHypervisorsV2_53(TestHypervisorsV28):
    COMPUTE_API_VERSION = "2.53"

    def test_list(self):
        self._test_list(cpu_info_type=dict, uuid_as_id=True)

    def test_search_with_details(self):
        # First find a hypervisor from the list to search on.
        hypervisors = self.client.hypervisors.list()
        # Now search for that hypervisor with details.
        hypervisor = hypervisors[0]
        hypervisors = self.client.hypervisors.search(
            hypervisor.hypervisor_hostname, detailed=True)
        self.assertEqual(1, len(hypervisors))
        hypervisor = hypervisors[0]
        # We know we got details if service is in the response.
        self.assertIsNotNone(hypervisor.service,
                             'Expected service in hypervisor: %s' % hypervisor)



# Copyright 2010 Jacob Kaplan-Moss
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from novaclient import base
from novaclient import exceptions
from novaclient.i18n import _


class Image(base.Resource):
    HUMAN_ID = True

    def __repr__(self):
        return "<Image: %s>" % self.name


class GlanceManager(base.Manager):
    """Use glance directly from service catalog.

    This is used to do name to id lookups for images and listing images for
    the --image-with option to the 'boot' command. Do not use it
    for anything else besides that. You have been warned.

    """

    resource_class = Image

    def find_image(self, name_or_id):
        """Find an image by name or id (user provided input)."""

        with self.alternate_service_type(
                'image', allowed_types=('image',)):
            # glance catalog entries are the unversioned endpoint, so
            # we need to jam a version bit in here.
            if uuidutils.is_uuid_like(name_or_id):
                # if it's a uuid, then just go collect the info and be
                # done with it.
                return self._get('/v2/images/%s' % name_or_id, None)
            else:
                matches = self._list('/v2/images?name=%s' % name_or_id,
                                     "images")
                num_matches = len(matches)
                if num_matches == 0:
                    msg = "No %s matching %s." % (
                        self.resource_class.__name__, name_or_id)
                    raise exceptions.NotFound(404, msg)
                elif num_matches > 1:
                    msg = (_("Multiple %(class)s matches found for "
                             "'%(name)s', use an ID to be more specific.") %
                           {'class': self.resource_class.__name__.lower(),
                            'name': name_or_id})
                    raise exceptions.NoUniqueMatch(msg)
                else:
                    matches[0].append_request_ids(matches.request_ids)
                    return matches[0]

    def find_images(self, names_or_ids):
        """Find multiple images by name or id (user provided input).

        :param names_or_ids: A list of strings to use to find images.
        :returns: novaclient.v2.images.Image objects for each images found
        :raises exceptions.NotFound: If one or more images is not found
        :raises exceptions.ClientException: If the image service returns any
                                            unexpected images.

        NOTE: This method always makes two calls to the image service, even if
        only one image is provided by ID and is returned in the first query.
        """
        with self.alternate_service_type(
                'image', allowed_types=('image',)):
            matches = self._list('/v2/images?id=in:%s' % ','.join(
                names_or_ids), 'images')
            matches.extend(self._list('/v2/images?names=in:%s' % ','.join(
                names_or_ids), 'images'))
            missed = (set(names_or_ids) -
                      set(m.name for m in matches) -
                      set(m.id for m in matches))
            if missed:
                msg = _("Unable to find image(s): %(images)s") % {
                    "images": ",".join(missed)}
                raise exceptions.NotFound(404, msg)
            for match in matches:
                match.append_request_ids(matches.request_ids)

            additional = []
            for i in matches:
                if i.name not in names_or_ids and i.id not in names_or_ids:
                    additional.append(i)
            if additional:
                msg = _('Additional images found in response')
                raise exceptions.ClientException(500, msg)
            return matches

    def list(self):
        """
        Get a detailed list of all images.

        :rtype: list of :class:`Image`
        """
        with self.alternate_service_type('image', allowed_types=('image',)):
            return self._list('/v2/images', 'images')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from tempest.lib import decorators

from novaclient.tests.functional import base
from novaclient.v2 import shell


@decorators.skip_because(bug="1675526")
class TestTriggerCrashDumpNovaClientV217(base.ProjectTestBase):
    """Functional tests for trigger crash dump"""

    COMPUTE_API_VERSION = "2.17"

    # It's a resource-consuming task to implement full-flow (up to getting
    # and reading a dump file) functional test for trigger-crash-dump.
    # We need to upload Ubuntu image for booting an instance based on it,
    # and to install kdump with its further configuring on this instance.
    # Here, the "light" version of functional test is proposed.
    # It's based on knowledge that trigger-crash-dump uses a NMI injection,
    # and when the 'trigger-crash-dump' operation is executed,
    # instance's kernel receives the NMI signal, and an appropriate
    # message will appear in the instance's log.

    # The server status must be ACTIVE, PAUSED, RESCUED, RESIZED or ERROR.
    # If not, the conflictingRequest(409) code is returned

    def _assert_nmi(self, server_id, timeout=60, poll_interval=1):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if 'trigger_crash_dump' in self.nova('instance-action-list %s ' %
                                                 server_id):
                break
            time.sleep(poll_interval)
        else:
            self.fail("Trigger crash dump hasn't been executed for server %s"
                      % server_id)

    def test_trigger_crash_dump_in_active_state(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('trigger-crash-dump %s ' % server.id)
        self._assert_nmi(server.id)

    def test_trigger_crash_dump_in_error_state(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('reset-state %s ' % server.id)
        shell._poll_for_status(
            self.client.servers.get, server.id,
            'active', ['error'])
        self.nova('trigger-crash-dump %s ' % server.id)
        self._assert_nmi(server.id)

    def test_trigger_crash_dump_in_paused_state(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('pause %s ' % server.id)
        shell._poll_for_status(
            self.client.servers.get, server.id,
            'active', ['paused'])
        self.nova('trigger-crash-dump %s ' % server.id)
        self._assert_nmi(server.id)

    def test_trigger_crash_dump_in_rescued_state(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('rescue %s ' % server.id)
        shell._poll_for_status(
            self.client.servers.get, server.id,
            'active', ['rescue'])
        self.wait_for_server_os_boot(server.id)
        self.nova('trigger-crash-dump %s ' % server.id)
        self._assert_nmi(server.id)

    def test_trigger_crash_dump_in_resized_state(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('resize %s %s' % (server.id, 'm1.small'))
        shell._poll_for_status(
            self.client.servers.get, server.id,
            'active', ['verify_resize'])
        self.nova('trigger-crash-dump %s ' % server.id)
        self._assert_nmi(server.id)

    def test_trigger_crash_dump_in_shutoff_state(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('stop %s ' % server.id)
        shell._poll_for_status(
            self.client.servers.get, server.id,
            'active', ['shutoff'])
        output = self.nova('trigger-crash-dump %s ' %
                           server.id, fail_ok=True, merge_stderr=True)
        self.assertIn("ERROR (Conflict): "
                      "Cannot 'trigger_crash_dump' instance %s "
                      "while it is in vm_state stopped (HTTP 409) " %
                      server.id, output)

    # If the specified server is locked, the conflictingRequest(409) code
    # is returned to a user without administrator privileges.
    def test_trigger_crash_dump_in_locked_state_admin(self):
        server = self._create_server()
        self.wait_for_server_os_boot(server.id)
        self.nova('lock %s ' % server.id)
        self.nova('trigger-crash-dump %s ' % server.id)
        self._assert_nmi(server.id)

    def test_trigger_crash_dump_in_locked_state_nonadmin(self):
        name = self.name_generate()
        server = self.another_nova('boot --flavor %s --image %s --poll %s' %
                                   (self.flavor.name, self.image.name, name))
        self.addCleanup(self.another_nova, 'delete', params=name)
        server_id = self._get_value_from_the_table(
            server, 'id')
        self.wait_for_server_os_boot(server_id)
        self.another_nova('lock %s ' % server_id)
        self.addCleanup(self.another_nova, 'unlock', params=name)
        output = self.another_nova('trigger-crash-dump %s ' %
                                   server_id, fail_ok=True, merge_stderr=True)
        # NOTE(mriedem): Depending on the version of the server you can get
        # different error messages back from this, so just assert that it's a
        # 409 either way.
        self.assertIn("ERROR (Conflict)", output)
